import torch


class FGM():
    def __init__(self, model):
        self.model = model
        self.backup = {}

    def attack(self, epsilon=1., emb_name='word_embeddings'):
        # emb_name这个参数要换成你模型中embedding的参数名
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0 and not torch.isnan(norm):
                    r_at = epsilon * param.grad / norm
                    param.data.add_(r_at)

    def restore(self, emb_name='word_embeddings'):
        # emb_name这个参数要换成你模型中embedding的参数名
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name: 
                assert name in self.backup
                param.data = self.backup[name]
        self.backup = {}


class PGD():
    def __init__(self, model):
        self.model = model
        self.emb_backup = {}
        self.grad_backup = {}

    def attack(self, epsilon=1., alpha=0.3, emb_name='word_embeddings', is_first_attack=False):
        # emb_name这个参数要换成你模型中embedding的参数名
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0 and not torch.isnan(norm):
                    r_at = alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, epsilon)

    def restore(self, emb_name='word_embeddings'):
        # emb_name这个参数要换成你模型中embedding的参数名
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name: 
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def project(self, param_name, param_data, epsilon):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                param.grad = self.grad_backup[name]



class FreeLB():
    def __init__(self, model, args, optimizer, base_model='xlm-roberta'):
        self.args = args
        self.model = model
        self.adv_K = self.args.adv_K
        self.adv_lr = self.args.adv_lr
        self.adv_max_norm = self.args.adv_max_norm
        self.adv_init_mag = self.args.adv_init_mag  # adv-training initialize with what magnitude, 即我们用多大的数值初始化delta
        self.adv_norm_type = self.args.adv_norm_type
        self.base_model = base_model
        self.optimizer = optimizer

    def attack(self, model, inputs):
        args = self.args
        input_ids = inputs['input_ids']
        #获取初始化时的embedding
        embeds_init = getattr(model, self.base_model).embeddings.word_embeddings(input_ids.to(args.device))

        if self.adv_init_mag > 0:   # 影响attack首步是基于原始梯度(delta=0)，还是对抗梯度(delta!=0)
            input_mask = inputs['attention_mask'].to(embeds_init)
            input_lengths = torch.sum(input_mask, 1)
            if self.adv_norm_type == "l2":
                delta = torch.zeros_like(embeds_init).uniform_(-1, 1) * input_mask.unsqueeze(2)
                dims = input_lengths * embeds_init.size(-1)
                mag = self.adv_init_mag / torch.sqrt(dims)
                delta = (delta * mag.view(-1, 1, 1)).detach()
        else:
            delta = torch.zeros_like(embeds_init)  # 扰动初始化
        # loss, logits = None, None


        for astep in range(self.adv_K):
            delta.requires_grad_()
            inputs['inputs_embeds'] = delta + embeds_init  # 累积一次扰动delta
            # inputs['input_ids'] = None
            loss, _ = model(input_ids=None,
                            attention_mask=inputs["attention_mask"].to(args.device),
                             token_type_ids=inputs["token_type_ids"].to(args.device),
                             labels=inputs["sl_labels"].to(args.device),
                            inputs_embeds=inputs["inputs_embeds"].to(args.device))

            loss = loss / self.adv_K

            loss.backward()

            if astep == self.adv_K - 1:
                # further updates on delta
                break

            delta_grad = delta.grad.clone().detach()  # 备份扰动的grad
            if self.adv_norm_type == "l2":
                denorm = torch.norm(delta_grad.view(delta_grad.size(0), -1), dim=1).view(-1, 1, 1)
                denorm = torch.clamp(denorm, min=1e-8)
                delta = (delta + self.adv_lr * delta_grad / denorm).detach()
                if self.adv_max_norm > 0:
                    delta_norm = torch.norm(delta.view(delta.size(0), -1).float(), p=2, dim=1).detach()
                    exceed_mask = (delta_norm > self.adv_max_norm).to(embeds_init)
                    reweights = (self.adv_max_norm / delta_norm * exceed_mask + (1 - exceed_mask)).view(-1, 1, 1)
                    delta = (delta * reweights).detach()
            else:
                raise ValueError("Norm type {} not specified.".format(self.adv_norm_type))

            embeds_init = getattr(model, self.base_model).embeddings.word_embeddings(input_ids.to(args.device))
        return loss
    
    
adversical_tricks = {'fgm': FGM, 'pgd': PGD, 'free_lb': FreeLB}


from typing import List
from ..data.doc import Entity
from pydantic import validate_arguments, constr  
    

def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == 'S':
        chunk_end = True
    # pred_label中可能出现这种情形
    if prev_tag == 'B' and tag == 'B':
        chunk_end = True
    if prev_tag == 'B' and tag == 'S':
        chunk_end = True
    if prev_tag == 'B' and tag == 'O':
        chunk_end = True
    if prev_tag == 'I' and tag == 'B':
        chunk_end = True
    if prev_tag == 'I' and tag == 'S':
        chunk_end = True
    if prev_tag == 'I' and tag == 'O':
        chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.
    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.
    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == 'B':
        chunk_start = True
    if tag == 'S':
        chunk_start = True

    if prev_tag == 'S' and tag == 'I':
        chunk_start = True
    if prev_tag == 'O' and tag == 'I':
        chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    return chunk_start

@validate_arguments
def get_chunks(seq: List[str]):
    """
    Gets chunks from bio(bios) tags.
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        get_chunks(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        tag = chunk[0]
        type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_
    return chunks

def convert_bio_to_entities(seq: List[str]) -> List[Entity]:
    """
    Gets entities from sequence.
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        get_entities(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """
    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]
    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        tag = chunk[0]
        type_ = chunk.split('-')[-1]

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_
    ents = [Entity(label=chunk[0], indices=[i for i in range(chunk[1], chunk[2]+1)]) for chunk in chunks]
    return ents


import logging
import warnings
from typing import List, Sequence
from lightning.pytorch.loggers import Logger, WandbLogger
from lightning.pytorch import LightningDataModule, LightningModule, Trainer, Callback
import rich.syntax
import rich.tree
from omegaconf import DictConfig, OmegaConf
from lightning.pytorch.utilities import rank_zero_only
import os


def get_logger(name=__name__) -> logging.Logger:
    """Initializes multi-GPU-friendly python command line logger."""

    logger = logging.getLogger(name)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in (
        "debug",
        "info",
        "warning",
        "error",
        "exception",
        "fatal",
        "critical",
    ):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def extras(config: DictConfig) -> None:
    """A couple of optional utilities, controlled by main config file:
    - disabling warnings
    - forcing debug friendly configuration
    - verifying experiment name is set when running in experiment mode

    Modifies DictConfig in place.

    Args:
        config (DictConfig): Configuration composed by Hydra.
    """

    log = get_logger(__name__)

    # disable python warnings if <config.ignore_warnings=True>
    if config.get("ignore_warnings"):
        log.info("Disabling python warnings! <config.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # verify experiment name is set when running in experiment mode
    if config.get("experiment_mode") and not config.get("name"):
        log.info(
            "Running in experiment mode without the experiment name specified! "
            "Use `python run.py mode=exp name=experiment_name`"
        )
        log.info("Exiting...")
        exit()

    # force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    # debuggers don't like GPUs and multiprocessing
    if config.trainer.get("fast_dev_run"):
        log.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        if config.trainer.get("gpus"):
            config.trainer.gpus = 0
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0


@rank_zero_only
def print_config(
    config: DictConfig,
    fields: Sequence[str] = (
        "trainer",
        "model",
        "datamodule",
        "callbacks",
        "logger",
        "test_after_training",
        "seed",
        "name",
    ),
    resolve: bool = True,
) -> None:
    """Prints content of DictConfig using Rich library and its tree structure.

    Args:
        config (DictConfig): Configuration composed by Hydra.
        fields (Sequence[str], optional): Determines which main fields from config will
        be printed and in what order.
        resolve (bool, optional): Whether to resolve reference fields of DictConfig.
    """

    style = "dim"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    for field in fields:
        branch = tree.add(field, style=style, guide_style=style)

        config_section = config.get(field)
        branch_content = str(config_section)
        if isinstance(config_section, DictConfig):
            branch_content = OmegaConf.to_yaml(config_section, resolve=resolve)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    rich.print(tree)

    with open("config_tree.log", "w") as fp:
        rich.print(tree, file=fp)


@rank_zero_only
def log_hyperparameters(
    config: DictConfig,
    model: LightningModule,
    datamodule: LightningDataModule,
    trainer: Trainer,
    callbacks: List[Callback],
    logger: List[Logger],
) -> None:
    """This method controls which parameters from Hydra config are saved by Lightning loggers.

    Additionaly saves:
        - number of model parameters
    """
    
    if not trainer.logger:
        return

    hparams = {}

    # choose which parts of hydra config will be saved to loggers
    hparams["trainer"] = config["trainer"]
    hparams["model"] = config["model"]
    hparams["datamodule"] = config["datamodule"]

    if "seed" in config:
        hparams["seed"] = config["seed"]
    if "callbacks" in config:
        hparams["callbacks"] = config["callbacks"]

    # save number of model parameters
    hparams["model/params/total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params/trainable"] = sum(
        p.numel() for p in model.parameters() if p.requires_grad
    )
    hparams["model/params/non_trainable"] = sum(
        p.numel() for p in model.parameters() if not p.requires_grad
    )

    # send hparams to all loggers
    trainer.logger.log_hyperparams(hparams)


def finish(
    config: DictConfig,
    model: LightningModule,
    datamodule: LightningDataModule,
    trainer: Trainer,
    callbacks: List[Callback],
    logger: List[Logger],
) -> None:
    """Makes sure everything closed properly."""

    # without this sweeps with wandb logger might crash!
    for lg in logger:
        if isinstance(lg, WandbLogger):
            import wandb

            wandb.finish()


def zip_all_files(dir,zipFile,pre_dir):
    """递归压缩文件夹下的所有文件
    参数:
    - dir: 要压缩的文件夹路径
    - zipFile: zipfile对象
    - pre_dir: 压缩文件根目录
    """
    for f in os.listdir(dir):
        absFile=os.path.join(dir,f) #子文件的绝对路径
        pre_d = os.path.join(pre_dir,f)
        if os.path.isdir(absFile): #判断是文件夹，继续深度读取。
            zipFile.write(absFile, pre_d) #在zip文件中创建文件夹
            zip_all_files(absFile,zipFile, pre_dir=pre_d) #递归操作
        else: #判断是普通文件，直接写到zip文件中。
            zipFile.write(absFile, pre_d)
    return 


from .ckpt_callbacks import LoadPLMStateDict, LoadModelStateDict
from pytorch_lightning.callbacks import ModelCheckpoint, ModelPruning, ModelSummary, RichModelSummary, EarlyStopping


__all__ = ["LoadPLMStateDict", 
           "LoadModelStateDict", 
           "ModelCheckpoint", 
           "ModelPruning", 
           "ModelSummary", 
           "RichModelSummary", 
           "EarlyStopping"]


import torch
from pytorch_lightning import LightningModule
from torchmetrics.classification import Accuracy
from typing import List, Any
from transformers import BertModel, BertConfig, BertTokenizer
import torch.nn.functional as F
import os


class BERTBiEncoder(LightningModule):
    '''双塔模型'''
    def __init__(
        self,
        hidden_size: int,
        lr: float,
        dropout: float,
        weight_decay: float,
        **kwargs
        ):
        super(BERTBiEncoder, self).__init__()
        self.save_hyperparameters(logger=False)
        self.bert = BertModel(self.hparams['bert_config'])
        self.dropout = torch.nn.Dropout(dropout)
        self.pooler = torch.nn.Linear(self.bert.config.hidden_size * 3, hidden_size)
        self.classifier = torch.nn.Linear(hidden_size, len(self.hparams.label2id))
        self.criterion = torch.nn.CrossEntropyLoss()
        self.train_acc = Accuracy(num_classes=len(self.hparams.label2id))
        self.val_acc = Accuracy(num_classes=len(self.hparams.label2id))
        self.test_acc = Accuracy(num_classes=len(self.hparams.label2id))
        self.tokenizer = self._init_tokenizer()

    def forward(self, inputs_a, inputs_b):
        encoded_a = self.dropout(torch.mean(self.bert(**inputs_a).last_hidden_state, dim=1))
        encoded_b = self.dropout(torch.mean(self.bert(**inputs_b).last_hidden_state, dim=1))
        abs_diff = torch.abs(encoded_a - encoded_b)
        concat = torch.cat((encoded_a, encoded_b, abs_diff), dim=-1)
        hidden = F.relu(self.pooler(concat))
        logits = self.classifier(hidden)
        return logits

    def shared_step(self, batch):
        inputs_a = batch['inputs_a']
        inputs_b = batch['inputs_b']
        labels = batch['label_ids']
        logits = self(inputs_a, inputs_b)
        loss = self.criterion(logits, labels)
        preds = torch.argmax(logits, dim=-1)
        return loss, preds, labels


    def training_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.train_acc(preds, labels)
        self.log('train/acc',self.train_acc, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def validation_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.val_acc(preds, labels)
        self.log('val/acc', self.val_acc, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def test_step(self, batch, batch_idx):
        loss, preds, labels = self.shared_step(batch)
        self.test_acc(preds, labels)
        self.log('test/acc', self.test_acc, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}

    def configure_optimizers(self):
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler] 


from .token_classification import BertTokenClassification, BertCRF
from .span_extraction import GlobalPointer
from .text_classification import BertForTextClassification
from .relation_extraction import GPLinkerForRelationExtraction, BLinkerForEntityRelationExtraction
from .text_multi_classification import BertTextMultiClassification
from .text_pair_classification import BERTBiEncoder, BERTCrossEncoder
from .text_pair_regression import SentenceBERT
from .prompt_relation_extraction import GPLinkerForPromptRelationExtraction
from .event_extraction import GPLinkerForEventExtraction, BiaffineForEventExtraction
from .entity_extraction import W2ForEntityExtraction, GlobalPointerForEntityExtraction, BiaffineForEntityExtraction, CRFForEntityExtraction
from .question_answering import PointerForQuestionAnswering


from typing import Any
from datasets import load_from_disk
from ..utils.make_datamodule import PLMBaseDataModule

class TextMultiClassification(PLMBaseDataModule):
    """多标签多分类的数据模块, 每个标签下有多重选择的情况"""
    def __init__(self,
                dataset: str,
                plm: str,
                max_length: int,
                batch_size: int,
                pin_memory: bool=False,
                num_workers: int=0,
                data_dir: str ='./datasets',
                plm_dir: str = './plms'):
        """多标签多分类数据模块, 每个标签下游多重选择

        Args:
            dataset (str): 数据集名称
            plm (str): 预训练模型名称
            max_length (int): 单文本最大长度
            batch_size (int): 批次大小
            pin_memory (bool, optional): _description_. Defaults to False.
            num_workers (int, optional): _description_. Defaults to 0.
            data_dir (str, optional): _description_. Defaults to './datasets'.
            plm_dir (str, optional): _description_. Defaults to './plms'.
        """
        super().__init__()
        self.save_hyperparameters()


    def transform(self, batch):
        raise NotImplementedError

    def setup(self, stage: str) -> None:
        """需要设置参数label2id, id2label, token2id, bert_config最后要对dataset设置transform"""
        raise NotImplementedError
    
    @property
    def dataset(self):
        return load_from_disk(self.hparams.data_dir + self.hparams.dataset)


import torch.nn as nn
import torch


class MultiDropout(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.dropout_0 = nn.Dropout(p=0)
        self.dropout_1 = nn.Dropout(p=0.1)
        self.dropout_2 = nn.Dropout(p=0.2)
        self.dropout_3 = nn.Dropout(p=0.3)
        self.dropout_4 = nn.Dropout(p=0.4)

        
    def forward(self, x):
        output_0 = self.dropout_0(x)
        output_1 = self.dropout_1(x)
        output_2 = self.dropout_2(x)
        output_3 = self.dropout_3(x)
        output_4 = self.dropout_4(x)
        return torch.mean(torch.stack([output_0,output_1,output_2,output_3,output_4], dim=0), dim=0)


import pytorch_lightning as pl
from ...layers import CRF, SimpleDense
from ...metrics.chunk import ChunkF1, get_entities
from transformers import BertModel, BertTokenizer
import torch.nn as nn
import os
import torch

class BertLstmCRF(pl.LightningModule):
    def __init__(self,
                 hidden_size: int,
                 lr: float,
                 weight_decay: float,
                 dropout: float,
                 **data_params):
        super().__init__()
        self.save_hyperparameters()
        self.label2id = data_params['label2id']
        self.id2label = {id: label for label, id in self.label2id.items()}
        self.tokenizer = self._init_tokenizer()

        self.bert = BertModel(data_params['bert_config'])
        self.rnn = nn.LSTM(
            input_size=self.bert.config.hidden_size, 
            hidden_size=hidden_size,
            num_layers=2,
            bidirectional=True,
            batch_first=True,
            dropout=dropout)

        self.classifier = SimpleDense(
            input_size=hidden_size*2, 
            hidden_size=hidden_size, 
            output_size=len(self.label2id))

        self.crf = CRF(len(self.label2id))
        self.dropout = nn.Dropout(dropout)

        self.train_f1 = ChunkF1()
        self.val_f1 = ChunkF1()
        self.test_f1 = ChunkF1()

        

    def forward(self, input_ids, token_type_ids, attention_mask, label_ids=None):
        last_hidden_state = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        hidden_state = self.dropout(last_hidden_state)
        rnn_out, hidden = self.rnn(hidden_state)
        emissions = self.classifier(rnn_out)
        mask = attention_mask.gt(0)
        if label_ids is not None :
            loss = self.crf(emissions, label_ids, mask=mask) * (-1)
            pred_ids= self.crf.decode(emissions, mask=mask)
            return loss, pred_ids
        else :
            return self.crf.decode(emissions, mask=mask)


    def shared_step(self, batch):
        inputs = batch['inputs']
        label_ids = batch['label_ids']
        #将label padding部分改为-1 
        label_ids[label_ids==self.hparams.label_pad_id] = -1
        loss, pred_ids = self(**inputs, label_ids=label_ids)
        pred_labels = []
        for ids in pred_ids:
            pred_labels.append([self.id2label[id] for id in ids])

        true_labels = []
        for i in range(len(label_ids)):
            indice = torch.where(label_ids[i] >= 0)
            ids = label_ids[i][indice].tolist()
            true_labels.append([self.id2label[id] for id in ids])
        return loss, pred_labels, true_labels

    
    def on_train_start(self) -> None:
        state_dict = torch.load(self.hparams.pretrained_dir + self.hparams.plm + '/pytorch_model.bin')
        self.bert.load_state_dict(state_dict)
        

    
    def training_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.train_f1(pred_labels, true_labels)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss':loss}


    def validation_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.val_f1(pred_labels, true_labels)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}

    
    def test_step(self, batch, batch_idx):
        loss, pred_labels, true_labels = self.shared_step(batch)
        self.test_f1(pred_labels, true_labels)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}



    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.bert.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.bert.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.rnn.named_parameters() if not any(nd in n for nd in no_decay)],
            'lr': self.hparams.lr * 100, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.rnn.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 100, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0},
            {'params': self.crf.parameters(), 'lr': self.hparams.lr * 500, 'weight_decay': self.hparams.weight_decay}
        ]
        self.optimizer = torch.optim.AdamW(grouped_parameters)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]


    def predict(self, text: str, device):
        inputs = self.tokenizer(
            text,
            add_special_tokens=True,
            return_tensors='pt')
        inputs.to(device)
        outputs = self(**inputs)
        labels = [self.id2label[id] for id in outputs[0]]
        ents = get_entities(seq=labels[1:-1])    #去掉cls sep 位置
        new_ents = []
        for ent in ents:
            new_ents.append([ent[1], ent[2], ent[0], text[ent[1]:ent[2]+1]])
        return new_ents


import torch.nn as nn
import torch
import torch.nn.functional as F 


class SkipGram(nn.Module):
    """word2vector模型skipgram模式的复现
    参考:
    - https://github.com/jasoncao11/nlp-notebook/blob/master/1-1.Word2Vec/model.py
    """
    def __init__(self,
                vocab_size,
                hidden_size) -> None:
        '''
        参数:
        - vocab_size: 词表大小
        - hidden_size: 词向量维度
        '''
        super().__init__()
        self.in_embedding = nn.Embedding(vocab_size, hidden_size)
        self.out_embedding = nn.Embedding(vocab_size, hidden_size)


    def forward(self, in_ids, pos_ids, neg_ids):
        """
        参数:
        - input_ids: 中心词的id, 也是词表的下标
        - pos_ids: 对应背景词的id
        - neg_ids: 负采样的词id
        """
        # batch_size * 1 -> batch_size * hidden_size *1
        in_embed = self.in_embedding(in_ids)
        in_embed = in_embed.unsqueeze(-1)
        # batch_size * (window*2) -> batch_size * (window*2) * hidden_size
        pos_embed = self.out_embedding(pos_ids)
        # batch_size * K -> batch_size * K * hidden_size
        neg_embed = self.out_embedding(neg_ids)
        pos_dot = torch.bmm(pos_embed, in_embed)
        pos_dot = pos_dot.squeeze(-1)

        # 负采样的词应该跟中心词距离越远
        neg_dot = torch.bmm(neg_embed, in_embed) * (-1)
        neg_dot = neg_dot.squeeze(-1)

        log_pos = F.logsigmoid(pos_dot)
        log_neg = F.logsigmoid(neg_dot)

        return -(log_pos + log_neg).mean()


if __name__ == "__main__":
    model = SkipGram(10, 20)
    model.eval()
    in_ids = torch.tensor([2])
    pos_ids = torch.tensor([[1,3]])
    neg_ids = torch.tensor([[4,5]])
    distance = model(in_ids, pos_ids, neg_ids)
    print(distance)








from ...utils.make_model import PLMBaseModel
from ...layers import LayerNorm, Biaffine
from ...metrics.entity import Entity, EntityF1
import torch.nn as nn
import torch.nn.functional as F
import torch
from typing import List
from collections import deque, defaultdict



class ConvolutionLayer(nn.Module):
    def __init__(self, input_size, channels, dilation, dropout=0.1):
        super(ConvolutionLayer, self).__init__()
        self.base = nn.Sequential(
            nn.Dropout2d(dropout),            #常用于图像
            nn.Conv2d(input_size, channels, kernel_size=1),   #将通道数由552改为96
            nn.GELU(),
        )

        self.convs = nn.ModuleList(
            [nn.Conv2d(channels, channels, kernel_size=3, groups=channels, dilation=d, padding=d) for d in dilation])

    def forward(self, x):
        x = x.permute(0, 3, 1, 2).contiguous()
        x = self.base(x)

        outputs = []
        for conv in self.convs:
            x = conv(x)
            x = F.gelu(x)
            outputs.append(x)
        outputs = torch.cat(outputs, dim=1)    #将三组卷积结果进行拼接
        outputs = outputs.permute(0, 2, 3, 1).contiguous()
        return outputs


class Biaffine(nn.Module):
    def __init__(self, n_in, n_out=1, bias_x=True, bias_y=True):
        super(Biaffine, self).__init__()

        self.n_in = n_in
        self.n_out = n_out
        self.bias_x = bias_x
        self.bias_y = bias_y
        weight = torch.zeros((n_out, n_in + int(bias_x), n_in + int(bias_y)))
        nn.init.xavier_normal_(weight)                                          #将权重变为正态分布
        self.weight = nn.Parameter(weight, requires_grad=True)

    def extra_repr(self):
        s = f"n_in={self.n_in}, n_out={self.n_out}"
        if self.bias_x:
            s += f", bias_x={self.bias_x}"
        if self.bias_y:
            s += f", bias_y={self.bias_y}"

        return s

    def forward(self, x, y):
        if self.bias_x:
            x = torch.cat((x, torch.ones_like(x[..., :1])), -1) #加偏执项
        if self.bias_y:
            y = torch.cat((y, torch.ones_like(y[..., :1])), -1)
        # [batch_size, n_out, seq_len, seq_len]
        s = torch.einsum('bxi,oij,byj->boxy', x, self.weight, y)  #矩阵点积
        # remove dim 1 if n_out == 1
        s = s.permute(0, 2, 3, 1)

        return s


class MLP(nn.Module):
    def __init__(self, n_in, n_out, dropout=0):
        super().__init__()

        self.linear = nn.Linear(n_in, n_out)
        self.activation = nn.GELU()
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = self.dropout(x)
        x = self.linear(x)
        x = self.activation(x)
        return x


class CoPredictor(nn.Module):
    def __init__(self, cls_num, hid_size, biaffine_size, channels, ffnn_hid_size, dropout=0):
        super().__init__()
        self.mlp1 = MLP(n_in=hid_size, n_out=biaffine_size, dropout=dropout)    #MLP就是一个线性变化器+一个激活函数+dropout
        self.mlp2 = MLP(n_in=hid_size, n_out=biaffine_size, dropout=dropout)
        self.biaffine = Biaffine(n_in=biaffine_size, n_out=cls_num, bias_x=True, bias_y=True)
        self.mlp_rel = MLP(channels, ffnn_hid_size, dropout=dropout)
        self.linear = nn.Linear(ffnn_hid_size, cls_num)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, y, z):           #x,y相同，都是每句话中每个字的特征,z是卷积得到的结果
        h = self.dropout(self.mlp1(x))
        t = self.dropout(self.mlp2(y))
        o1 = self.biaffine(h, t)

        z = self.dropout(self.mlp_rel(z))
        o2 = self.linear(z)
        return o1 + o2
    

class W2ForEntityExtraction(PLMBaseModel):
    """w2ner 统一解决嵌套非连续实体, 模型复现
    
    修改:
        - 去除了lstm层
        
    参考: 
        - https://github.com/ljynlp/W2NER
        
    参数:
        - lr (float): 学习率
        - scheduler (str, optional): 学习率调度器名称. Defaults to 'linear_warmup_step'.
        - weight_decay (float, optional): 权重衰减. Defaults to 0.01.
        - conv_hidden_size (int, optional): 卷积层维度. Defaults to 96.
        - dilation (List[int], optional): . Defaults to [1,2,3].
        - dist_emb_size (int, optional): 距离编码维度. Defaults to 20.
        - reg_emb_size (int, optional): 区域编码维度. Defaults to 20.
        - biaffine_size (int, optional): 放射变换维度. Defaults to 512.
        - dropout (float, optional): dropout比率. Defaults to 0.2.
        - ffnn_hidden_size (int, optional): . Defaults to 288.
        - **kwargs : datamodule.hparams.
    """
    def __init__(self,
                 lr: float =3e-5,
                 scheduler: str = 'linear_warmup',
                 weight_decay: float = 0.01,
                 conv_hidden_size: int = 96,
                 dilation: List[int] = [1,2,3],
                 dist_emb_size: int = 20,
                 reg_emb_size: int = 20,
                 biaffine_size: int = 512,
                 dropout: float = 0.2,
                 ffnn_hidden_size: int = 288,
                 **kwargs):      
        super().__init__()
        
        self.plm = self.get_plm_architecture()

        self.distance_embeddings = nn.Embedding(20, self.hparams.dist_emb_size)
        self.region_embeddings = nn.Embedding(3, self.hparams.reg_emb_size)

        self.dropout = nn.Dropout(self.hparams.dropout)

        self.cln = LayerNorm(hidden_size=self.plm.config.hidden_size, 
                             conditional_size=self.plm.config.hidden_size)
        
        conv_input_size = self.hparams.dist_emb_size + self.hparams.reg_emb_size + self.plm.config.hidden_size
        self.convLayer = ConvolutionLayer(input_size=conv_input_size,
                                          channels=self.hparams.conv_hidden_size,
                                          dilation=self.hparams.dilation,
                                          dropout=self.hparams.dropout)

        self.predictor = CoPredictor(cls_num=len(self.hparams.label2id), 
                                     hid_size=self.plm.config.hidden_size, 
                                     biaffine_size=self.hparams.biaffine_size,
                                     channels=self.hparams.conv_hidden_size * len(self.hparams.dilation), 
                                     ffnn_hid_size=self.hparams.ffnn_hidden_size,
                                     dropout=self.hparams.dropout)

        self.criterion = nn.CrossEntropyLoss(reduction='none')
        self.train_metric = EntityF1()
        self.val_metric = EntityF1()


    def forward(self, input_ids, token_type_ids, attention_mask, grid_mask, distance_ids):
        # 将bert的输出取后四层的平均 -> b, l, h
        hiddens = self.plm(input_ids, token_type_ids, attention_mask, output_hidden_states=True).hidden_states
        hidden = torch.stack(hiddens[-4:], dim=-1).mean(-1) 
        hidden = self.dropout(hidden)
        # 经过cln -> b, l, l, h
        cln = self.cln(hidden.unsqueeze(2), hidden)
        # 将distance, region 和 cln输出 拼接 -> b,l,l,h_concat
        dis_emb = self.distance_embeddings(distance_ids)
        region_ids = self.get_region_ids(grid_mask)
        reg_emb = self.region_embeddings(region_ids)
        conv_inputs = torch.cat([dis_emb, reg_emb, cln], dim=-1)
        # 经过卷积层 -> b, l, l, h_ffnn
        conv_inputs = torch.masked_fill(conv_inputs, grid_mask.eq(0).unsqueeze(-1), 0.0) #grid_mask2d根据句子长度对句子设置padding，即将conv_inputs中padding的位置用0填补
        conv_outputs = self.convLayer(conv_inputs)                      #将3个 Dilated Convolution的结果拼接输出
        conv_outputs = torch.masked_fill(conv_outputs, grid_mask.eq(0).unsqueeze(-1), 0.0)
        # 输入预测层 -> b, l, l, num_labels
        logits = self.predictor(hidden, hidden, conv_outputs)

        return logits


    def get_region_ids(self, grid_mask):
        # tril_mask: batch seq seq
        tril_mask = torch.tril(grid_mask)
        region_ids = tril_mask + grid_mask
        return region_ids


    def get_grid_mask(self, attention_mask):
        grid_mask = torch.transpose(attention_mask.unsqueeze(1), 1, 2) * attention_mask.unsqueeze(1)
        return grid_mask


    def extract_ents(self, label_ids, attention_mask):
        class Node:
            def __init__(self):
                self.THW = []                # [(tail, type)]
                self.NNW = defaultdict(set)
        q = deque()
        ents = []
        length = torch.sum(attention_mask, dim=-1)
        for instance, l, in zip(label_ids, length):
            l = l.item()
            nodes = [Node() for _ in range(l)]
            predicts = []
            for cur in reversed(range(l)):
                heads = []
                for pre in range(cur+1):
                    if instance[cur, pre] > 1: 
                            nodes[pre].THW.append((cur, instance[cur, pre]))
                            heads.append(pre)
                        # NNW
                    if pre < cur and instance[pre, cur] == 1:
                        # cur node
                        for head in heads:
                            nodes[pre].NNW[(head,cur)].add(cur)
                        # post nodes
                        for head,tail in nodes[cur].NNW.keys():
                            if tail >= cur and head <= pre:
                                nodes[pre].NNW[(head,tail)].add(cur)
                for tail,type_id in nodes[cur].THW:
                        if cur == tail:
                            predicts.append(([cur], type_id))
                            continue
                        q.clear()
                        q.append([cur])
                        while len(q) > 0:
                            chains = q.pop()
                            for idx in nodes[chains[-1]].NNW[(cur,tail)]:
                                if idx == tail:
                                    predicts.append((chains + [idx], type_id))
                                else:
                                    q.append(chains + [idx])
            preds = [Entity(label=l.item(), indexes=tuple(i)) for i,l in predicts]
            ents.append(set(preds))
        return ents


    def step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        distance_ids = batch['distance_ids']
        label_ids = batch['label_ids']
        grid_mask = self.get_grid_mask(attention_mask)
        logits = self(input_ids, token_type_ids, attention_mask, grid_mask, distance_ids)
        loss = self.criterion(logits.permute(0, 3, 1, 2), label_ids)
        loss = torch.sum(loss * grid_mask) / torch.sum(grid_mask) 
        return logits, loss

    
    def training_step(self, batch, batch_idx):
        logits, loss = self.step(batch)
        return {'loss':loss}


    def validation_step(self, batch, batch_idx):
        targs = batch['label_ids']
        attention_mask = batch['attention_mask']
        targs = self.extract_ents(targs, attention_mask)
        logits, loss = self.step(batch)
        preds = logits.argmax(dim=-1)
        preds = self.extract_ents(preds, attention_mask)
        self.val_metric(preds, targs)
        self.log('val/f1', self.val_metric, on_epoch=True, prog_bar=True)


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr,'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr,'weight_decay': 0.0},
            
            {'params': [p for n, p in self.convLayer.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr,'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.convLayer.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr,'weight_decay': 0.0},
            
            {'params': [p for n, p in self.cln.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 10,'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.cln.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 10,'weight_decay': 0.0},
            
            {'params': [p for n, p in self.predictor.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 20,'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.predictor.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 20,'weight_decay': 0.0},
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def predict(self, text):
        pass


from functools import lru_cache
from typing import Optional,  Tuple, List, Union
from ..utils.make_datamodule import PLMBaseDataModule
import torch


class TextPairClassificationDataModule(PLMBaseDataModule):
    '''句子对数据模块,用来构建pytorch_lightning的数据模块.
    dataset_example:
        {'text_a': '肺结核','text_b': '关节炎','label': 不是一种病}
    '''
    def __init__(self,
                dataset: str,
                batch_size: int,
                plm: str,
                **kwargs): 
        """参数:
        - dataset: 数据集名称,feature 必须包含 text_a, text_b, label
        - plm: 预训练模型名称
        - batch_size: 批大小
        - return_pair: 是否以文本对为输入
        """ 
        super().__init__()


    @property
    @lru_cache()
    def label2id(self):
        set_labels = sorted(set([label for label in self.dataset['train']['label']]))
        return {label: i for i, label in enumerate(set_labels)}
    
    @property
    def id2label(self) -> dict:
        return {i:l for l,i in self.label2id.items()}

        
    def setup(self, stage: str):
        self.hparams['label2id'] = self.label2id
        self.hparams['id2label'] = self.id2label
        

    def cross_transform(self, examples):
        batch_text_a = examples['text_a']
        batch_text_b = examples['text_b']
        batch_labels = examples['label']
        a_max_length = self.get_batch_max_length(batch_text=batch_text_a)
        b_max_length = self.get_batch_max_length(batch_text=batch_text_b)
        max_length = a_max_length + b_max_length 
        max_length = min(max_length, 512)
        batch_inputs = self.tokenizer(batch_text_a,
                                      batch_text_b, 
                                      padding=True, 
                                      max_length=max_length, 
                                      truncation=True,
                                      return_tensors='pt')
        batch_label_ids = []
        for i in range(len(batch_text_a)):
            batch_label_ids.append(self.hparams['label2id'][batch_labels[i]])
        batch_label_ids = torch.tensor(batch_label_ids, dtype=torch.long)
        batch_inputs['label_ids'] = batch_label_ids
        return batch_inputs
        
            
        
    def bi_transform(self, examples):
        batch_text_a = examples['text_a']
        batch_text_b = examples['text_b']
        batch_labels = examples['label']
        batch = {'inputs_a': [], 'inputs_b': [], 'label_ids':[]}
        batch_cross = {'inputs': [], 'label_ids':[]}
        for i  in range(len(batch_text_a)):
                inputs_a= self.tokenizer(batch_text_a[i], 
                                        padding='max_length', 
                                        max_length=self.hparams.max_length+2, 
                                        truncation=True)
                inputs_a = dict(zip(inputs_a.keys(), map(torch.tensor, inputs_a.values())))
                batch['inputs_a'].append(inputs_a)
                inputs_b = self.tokenizer(batch_text_b[i],
                                        padding='max_length', 
                                        max_length=self.hparams.max_length, 
                                        truncation=True)
                inputs_b = dict(zip(inputs_b.keys(), map(torch.tensor, inputs_b.values())))
                batch['inputs_b'].append(inputs_b)
                batch['label_ids'].append(self.hparams['label2id'][batch_labels[i]])
        return  batch


import torch
from torch import nn
from torch.nn import Module
from ..embedding import SinusoidalPositionEmbedding, RoPEPositionEncoding


class GlobalPointer(Module):
    """全局指针模块,将序列的每个(start, end)作为整体来进行判断
    
    更改
    - 原版模型用矩阵点乘来得到span的表征,这里增加了叉乘,相加和拼接的方法,用来做对比实验
    """
    def __init__(self,
                 input_size: int, 
                 hidden_size: int,
                 output_size: int,  
                 add_rope: bool =True,
                 tril_mask: bool = True,
                 span_get_type: str = 'dot'):
        """

        Args:
            input_size (int): 输入维度大小一般为bert输出的hidden size
            hidden_size (int): 内部隐层的维度大小
            output_size (int): 一般为分类的span类型
            add_rope (bool, optional): 是否添加RoPE. Defaults to True.
            tril_mask (bool, optional): 是否遮掩下三角. Defaults to True.
            span_get_type (str, optional): span表征的获得方式.可以为'dot' 'element-product' 'concat' 'element-add',默认'dot',实验dot也是最好.
        """
        super(GlobalPointer, self).__init__()
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.add_rope = add_rope
        self.tril_mask = tril_mask
        self.span_get_type = span_get_type
        if self.span_get_type == 'dot':
        # 加了relu激活函数在hidden_size很小(32)的情况下,损失不下降,所以去掉了ReLU激活函数
        # self.q = nn.Sequential(nn.Linear(input_size, hidden_size * output_size), nn.ReLU())
        # self.k = nn.Sequential(nn.Linear(input_size, hidden_size * output_size), nn.ReLU())
            self.start = nn.Linear(input_size, hidden_size * output_size)
            self.end = nn.Linear(input_size, hidden_size * output_size)
        if self.span_get_type == 'element-product' or self.span_get_type == 'element-add':
            self.start = nn.Linear(input_size, hidden_size)
            self.end = nn.Linear(input_size, hidden_size) 
            self.span = nn.Linear(hidden_size, output_size)
        if self.span_get_type == 'concat':
            self.start = nn.Linear(input_size, hidden_size)
            self.end = nn.Linear(input_size, hidden_size) 
            self.span =  nn.Linear(hidden_size*2, output_size)


    def forward(self, inputs, mask=None):
        if self.span_get_type == 'dot':
            batch_size, seq_length, input_size = inputs.shape
            start_logits = self.start(inputs).reshape(batch_size, seq_length, self.output_size, self.hidden_size)
            end_logits = self.end(inputs).reshape(batch_size, seq_length, self.output_size, self.hidden_size)
            # 分出qw和kw
            # RoPE编码
            if self.add_rope:
                pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
                cos_pos = pos[..., None, 1::2].repeat(1, 1, 1, 2)
                sin_pos = pos[..., None, ::2].repeat(1, 1, 1, 2)
                start_logits2 = torch.stack([-start_logits[..., 1::2], start_logits[..., ::2]], 4)
                start_logits2 = torch.reshape(start_logits2, start_logits.shape)
                start_logits = start_logits * cos_pos + start_logits2 * sin_pos
                end_logits2 = torch.stack([-end_logits[..., 1::2], end_logits[..., ::2]], 4)
                end_logits2 = torch.reshape(end_logits2, end_logits.shape)
                end_logits = end_logits * cos_pos + end_logits2 * sin_pos
            # 计算内积
            logits = torch.einsum('bmhd , bnhd -> bhmn', start_logits, end_logits)
            logits = logits / self.hidden_size ** 0.5
        elif self.span_get_type == 'element-product' or self.span_get_type == 'element-add':
            start_logits = self.start(inputs)
            end_logits = self.end(inputs)
            if self.add_rope:
                pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
                cos_pos = pos[..., 1::2].repeat(1, 1, 2)
                sin_pos = pos[..., ::2].repeat(1, 1, 2)
                start2 = torch.stack([-start_logits[..., 1::2], start_logits[..., ::2]], 3)
                start2 = torch.reshape(start2, start_logits.shape)
                start_logits = start_logits * cos_pos + start2 * sin_pos
                end2 = torch.stack([-end_logits[..., 1::2], end_logits[..., ::2]], 3)
                end2 = torch.reshape(end2, end_logits.shape)
                end_logits = end_logits * cos_pos + end2 * sin_pos
            start_logits = start_logits.unsqueeze(1)
            end_logits = end_logits.unsqueeze(2)
            if self.span_get_type == 'element-product':
                logits = self.span(start_logits * end_logits).permute(0,3,2,1) # batch output seq seq
            elif self.span_get_type == 'element-add':
                logits = self.span(start_logits + end_logits).permute(0,3,2,1)
                
        elif self.span_get_type == 'concat':
            batch_size, seq_len, input_size = inputs.size()
            # head: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,1,1, 2,2,2, 3,3,3
            start_logits = self.start(inputs)
            end_logits = self.end(inputs)
            if self.add_rope:
                pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
                cos_pos = pos[..., 1::2].repeat(1, 1, 2)
                sin_pos = pos[..., ::2].repeat(1, 1, 2)
                start2 = torch.stack([-start_logits[..., 1::2], start_logits[..., ::2]], 3)
                start2 = torch.reshape(start2, start_logits.shape)
                start_logits = start_logits * cos_pos + start2 * sin_pos
                end2 = torch.stack([-end_logits[..., 1::2], end_logits[..., ::2]], 3)
                end2 = torch.reshape(end2, end_logits.shape)
                end_logits = end_logits * cos_pos + end2 * sin_pos
            start_logits = start_logits.unsqueeze(2).expand(batch_size, seq_len, seq_len, self.hidden_size).reshape(batch_size, seq_len*seq_len, self.hidden_size)
            # end: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,2,3, 1,2,3, 1,2,3
            end_logits = end_logits.repeat(1, seq_len, 1)
            pairs = torch.cat([start_logits, end_logits], dim=-1)
            logits = self.span(pairs).reshape(batch_size, self.output_size, seq_len, seq_len)
            
        # padding mask
        if mask is not None:
            batch_size = inputs.size()[0]
            seq_len = inputs.size()[1]
            pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
            logits = logits*pad_mask - (1-pad_mask)*1e12

        # 排除下三角
        if self.tril_mask:
            mask = torch.tril(torch.ones_like(logits), -1) 
            logits = logits - mask * 1e12

        # scale返回
        return logits



class EfficientGlobalPointer(Module):
    """全局指针模块
    将序列的每个(start, end)作为整体来进行判断
    https://kexue.fm/archives/8877
    """
    def __init__(self, 
                 input_size: int, 
                 hidden_size: int,
                 output_size: int,
                 add_rope: bool =True,
                 tril_mask: bool =True,
                 use_bias: bool = True,
                 max_length: int = 512):
        super(EfficientGlobalPointer, self).__init__()
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.add_rope = add_rope
        self.tril_mask = tril_mask
        self.input_size = input_size
        self.linear_1 = nn.Linear(input_size, hidden_size * 2, bias=use_bias)
        self.linear_2 = nn.Linear(hidden_size * 2, output_size * 2, bias=use_bias)
        if self.add_rope:
            self.pe = RoPEPositionEncoding(max_position=max_length, embedding_size=hidden_size)

    def forward(self, inputs, mask=None):
        inputs = self.linear_1(inputs)
        qw, kw = inputs[..., ::2], inputs[..., 1::2]
        # RoPE编码
        if self.add_rope:
            qw = self.pe(qw)
            kw = self.pe(kw)
        logits = torch.einsum('bmd , bnd -> bmn', qw, kw) / self.hidden_size ** 0.5
        # bias = self.linear_2(inputs)
        # bias = torch.stack(torch.chunk(bias, self.output_size, dim=-1), dim=-2).transpose(1,2) #[btz, heads, seq_len, 2]
        # logits = logits.unsqueeze(1) + bias[..., :1] + bias[..., 1:].transpose(2, 3)
        bias = self.linear_2(inputs).transpose(1, 2) / 2  #'bnh->bhn'
        logits = logits[:, None] + bias[:, ::2, None] + bias[:, 1::2, :, None]
        
        # padding mask
        if mask is not None:
            batch_size = inputs.size()[0]
            seq_len = inputs.size()[1]
            pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
            logits = logits*pad_mask - (1-pad_mask)*1e12

        # tril mask
        if self.tril_mask:
            t_mask = torch.tril(torch.ones_like(logits), -1) 
            logits = logits - t_mask * 1e12

        return logits


import torch.nn as nn
import torch.nn.functional as F
import torch
import math

# 乘性的注意力算法
def scaled_mul_self_attention(query, key, value, mask=None):
    """
    # 为什么scaled(也就是除以sqrt(d_k))？: 
    - 1. 维度升高使得乘性注意力机制的方差变大
    - 2. 进而出现极大值使得softmax梯度消失
    - 3. 通过scale控制方差,进而稳定梯度流, 防止梯度爆炸
    - https://www.zhihu.com/question/339723385
    - notebooks/transformers/attention.ipynb
    """
    d_scale = query.size(-1)
    logits = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_scale)
    if mask is not None:
        logits = logits.masked_fill(mask == 0, -1e9)
    scores = F.softmax(logits, dim=-1)
    output = torch.matmul(scores, value)
    return output, scores

# 存放注意力算法
attentions = {
    'scaled_mul_self_attention': scaled_mul_self_attention
    }


class MultiHeadAttentionLayer(nn.Module):
    def __init__(self, 
                hidden_size:int = 768,
                num_attention_heads: int = 8,
                attention_probs_dropout_prob: float = 0.1,
                return_attention_scores: bool = False,
                attention_scale: bool= True,
                bias:bool =True,
                **kwargs):

        """多头注意力机制
        参数:
        - hidden_size: 隐层维度
        - num_attention_heads: 注意力头的数量
        - attention_probs_dropout_prob: 注意力机制后dropout的概率
        - return_attention_scores: 是否返回注意力得分
        - attention_scale: 是否对注意力值缩放,默认True
        - bias: 

        参考: 
        - https://github.com/mmmwhy/pure_attention/blob/v0.0.22/pure_attention/backbone_bert/bert_layer.py
        - https://github.com/Tongjilibo/bert4torch/blob/a0db5a59a1ec2ec4820c4d055ceef463ce4e5d28/bert4torch/layers.py#L1
        """
        super().__init__()
        assert hidden_size % num_attention_heads == 0, "隐层维度不能被多头注意力头数整除"
        self.hidden_size = hidden_size
        self.num_attention_heads = num_attention_heads 
        self.attention_head_size  = int(hidden_size / num_attention_heads)
        self.attention_scale = attention_scale
        self.return_attention_scores = return_attention_scores
        self.query = nn.Linear(hidden_size, hidden_size)
        self.key = nn.Linear(hidden_size, hidden_size)
        self.value= nn.Linear(hidden_size, hidden_size)
        self.dropout = nn.Dropout(attention_probs_dropout_prob)

        self.a_bias, self.p_bias = kwargs.get('a_bias'), kwargs.get('p_bias')

        if self.p_bias == 'typical_relative':
            pass

    def forward(self, query, key, value, attention_mask=None, head_mask=None):
        mixed_query_layer = self.query(query)
        mixed_key_layer = self.key(key)
        mixed_value_layer = self.value(value)
        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(mixed_key_layer)
        value_layer = self.transpose_for_scores(mixed_value_layer)
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2)) / math.sqrt(self.attention_head_size)
        if attention_mask is not None:
            attention_scores += attention_mask
        attention_probs = nn.Softmax(dim=-1)(attention_scores)
        if head_mask is not None:
            attention_probs += head_mask
        context_layer = torch.matmul(attention_scores, value_layer)
        # view操作需要内存连续, permute和transpose之后张量在内存中不再连续所以要加上contiguous
        context_layer = context_layer.permute(0,2,1,3).contiguous()
        new_context_layer_size = context_layer.size()[:-2] + (self.hidden_size)
        context_layer = context_layer.view(*new_context_layer_size)
        outputs = (context_layer, attention_scores) if self.return_attention_scores else (context_layer,)
        return outputs

    def transpose_for_scores(self, x):
        new_x_size = x.size()[:-1] + (self.num_attention_heads, self.attention_head_size)
        x = x.view(*new_x_size)
        x = x.permute(0, 2, 1, 3)
        return x         


import pytorch_lightning as pl
from transformers import BertModel, BertTokenizer
import torch
import torch.nn.functional as F
import os
from ...metrics.chunk import ChunkF1
from ...layers import SimpleDense


class BertTokenClassification(pl.LightningModule):
    def __init__(self,
                 hidden_size: int,
                 lr: float,
                 weight_decay: float,
                 dropout:float ,
                 **data_params):
        super().__init__()
        self.save_hyperparameters()
        self.label2id = data_params['label2id']
        self.id2label = {id: label for label, id in self.label2id.items()}
        

        self.bert = BertModel(data_params['bert_config'])
        self.classifier = SimpleDense(self.bert.config.hidden_size, hidden_size, len(self.label2id))
        self.criterion = torch.nn.CrossEntropyLoss()
        self.train_f1 = ChunkF1()
        self.val_f1 = ChunkF1()
        self.test_f1 = ChunkF1()

        
        
    
    def forward(self, input_ids, token_type_ids, attention_mask):
        bert_hidden_state = self.bert(input_ids, token_type_ids, attention_mask).last_hidden_state
        logits = self.classifier(bert_hidden_state)
        return logits

    def shared_step(self, batch):
        inputs = batch['inputs']
        label_ids = batch['label_ids']
        logits = self(**inputs)
        mask = torch.gt(label_ids, -1)
        loss = self.criterion(logits.view(-1, len(self.label2id))[mask.view(-1)], label_ids.view(-1)[mask.view(-1)])
        pred_ids = torch.argmax(logits, dim=-1)
        labels = []
        pred_labels = []
        for i in range(len(label_ids)):
            indice = torch.where(label_ids[i] >= 0)
            ids = label_ids[i][indice].tolist()
            labels.append([self.id2label[id] for id in ids])
            pred = pred_ids[i][indice].tolist()
            pred_labels.append([self.id2label[id] for id in pred])
        return loss, labels, pred_labels


    def on_train_start(self) -> None:
        state_dict = torch.load(self.hparams.pretrained_dir + self.hparams.plm + '/pytorch_model.bin')
        self.bert.load_state_dict(state_dict)


    def training_step(self, batch, batch_idx):
        loss, labels, pred_labels = self.shared_step(batch)
        self.train_f1(true=labels, pred=pred_labels)
        self.log('train/f1',self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss':loss}


    def validation_step(self, batch, batch_idx):
        loss, labels, pred_labels = self.shared_step(batch)
        self.val_f1(true=labels, pred=pred_labels)
        self.log('val/f1',self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}


    def test_step(self, batch, batch_idx):
        loss, labels, pred_labels = self.shared_step(batch)
        self.test_f1(true=labels, pred=pred_labels)
        self.log('test/f1',self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}

    def configure_optimizers(self):
        self.optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay)
        self.scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, lambda epoch: 1.0 / (epoch + 1.0))
        return [self.optimizer], [self.scheduler]







import torch
import torch.nn as nn
from ..embedding import SinusoidalPositionEmbedding


class Biaffine(nn.Module):
    def __init__(self, input_size, output_size=1, bias_x=True, bias_y=True):
        super(Biaffine, self).__init__()

        self.input_size = input_size
        self.output_size = output_size
        self.bias_x = bias_x
        self.bias_y = bias_y
        weight = torch.zeros((output_size, input_size + int(bias_x), input_size + int(bias_y)))
        nn.init.xavier_normal_(weight)                                          #将权重变为正态分布
        self.weight = nn.Parameter(weight, requires_grad=True)

    def extra_repr(self):
        s = f"input_size={self.input_size}, output_size={self.output_size}"
        if self.bias_x:
            s += f", bias_x={self.bias_x}"
        if self.bias_y:
            s += f", bias_y={self.bias_y}"

        return s

    def forward(self, x, y):
        if self.bias_x:
            x = torch.cat((x, torch.ones_like(x[..., :1])), -1) #加偏执项
        if self.bias_y:
            y = torch.cat((y, torch.ones_like(y[..., :1])), -1)
        # [batch_size, output_size, seq_len, seq_len]
        s = torch.einsum('bxi,oij,byj->boxy', x, self.weight, y)  #矩阵点积
        # remove dim 1 if output_size == 1
        s = s.permute(0, 2, 3, 1)

        return s  
    
    
class BiaffineSpanClassifier(nn.Module):
    def __init__(self, 
                 input_size: int, 
                 output_size: int, 
                 hidden_size: int,
                 add_rope: bool = True,
                 tril_mask: bool = True):
        super().__init__()

        self.input_size = input_size
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.tril_mask = tril_mask
        self.add_rope = add_rope
        self.start_repr = nn.Linear(input_size, hidden_size)
        self.end_repr = nn.Linear(input_size, hidden_size)
        weight = torch.zeros((output_size, hidden_size, hidden_size))
        nn.init.xavier_normal_(weight)                                          #将权重变为正态分布
        self.weight = nn.Parameter(weight, requires_grad=True)


    def forward(self, x, mask=None):
        """将序列的表示(batch,seq,hidden),转换为token pair的表示

        Args:
            x : [batch_size, seq_len, hidden_size]
            mask : [batch_size, seq_len]

        Returns:
            logits: [batch_size, output_size, seq_len, seq_len]
        """
        start_logits = self.start_repr(x)
        end_logits = self.end_repr(x)
        if self.add_rope:
            pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(x)
            cos_pos = pos[..., 1::2].repeat(1, 1, 2)
            sin_pos = pos[..., ::2].repeat(1, 1, 2)
            start2 = torch.stack([-start_logits[..., 1::2], start_logits[..., ::2]], 3)
            start2 = torch.reshape(start2, start_logits.shape)
            start_logits = start_logits * cos_pos + start2 * sin_pos
            end2 = torch.stack([-end_logits[..., 1::2], end_logits[..., ::2]], 3)
            end2 = torch.reshape(end2, end_logits.shape)
            end_logits = end_logits * cos_pos + end2 * sin_pos

        # [batch_size, output_size, seq_len, seq_len]
        span_logits = torch.einsum('bxi,oij,byj->boxy', start_logits, self.weight, end_logits)  #矩阵点积            
        
        
        if mask is not None:
            batch_size = x.size()[0]
            seq_len = x.size()[1]
            pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
            span_logits = span_logits*pad_mask - (1-pad_mask) * 1e12
        
        if self.tril_mask:
            mask_tril = torch.tril(torch.ones_like(span_logits), diagonal=-1)
            span_logits = span_logits - mask_tril * 1e12
            
        return span_logits
    
    
class EfficientBiaffineSpanClassifier(nn.Module):
    """更加高效的利用biaffine机制做span分类,想法来源于EfficientGlobalPointer
    
    - 输入: batch_size, seq_length, hidden_size
    - 输出: batch_size, seq_length, seq_length, output_size(num_class)
    """
    def __init__(self, 
                 input_size: int, 
                 hidden_size: int,
                 output_size: int,
                 add_rope: bool =True,
                 tril_mask: bool =True):
        super().__init__()
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.RoPE = add_rope
        self.tril_mask = tril_mask
        self.input_size = input_size
        self.head = nn.Linear(input_size, hidden_size)
        self.tail = nn.Linear(input_size, hidden_size)
        weight = torch.zeros(output_size, hidden_size, hidden_size)
        nn.init.xavier_normal_(weight)
        self.cls = nn.Parameter(weight, requires_grad=True)

    def forward(self, inputs, mask=None):
        qw = self.head(inputs)
        kw = self.tail(inputs)
        # RoPE编码
        if self.RoPE:
            pos = SinusoidalPositionEmbedding(self.hidden_size, 'zero')(inputs)
            cos_pos = pos[...,1::2].repeat(1,1,2)
            sin_pos = pos[...,::2].repeat(1,1,2)
            qw2 = torch.stack([-qw[..., 1::2], qw[..., ::2]], 3)
            qw2 = torch.reshape(qw2, qw.shape)
            qw = qw * cos_pos + qw2 * sin_pos
            kw2 = torch.stack([-kw[..., 1::2], kw[..., ::2]], 3)
            kw2 = torch.reshape(kw2, kw.shape)
            kw = kw * cos_pos + kw2 * sin_pos
        # 得到每个span的边界分数 -> batch, 1, seqlen, seqlen
        span = torch.einsum('bmd , bnd -> bmn', qw, kw) / self.hidden_size**0.5
        span = span.unsqueeze(1)
        
        # 得到每个span的类别分数 基于双仿射机制得到类别分数 -> batch, class, seqlen, seqlen
        cls = torch.einsum('bxi,oij,byj->boxy', qw, self.cls, kw)
        
        # 边界得分与类别得分相加 除2为了模型更容易学习
        logits = (span + cls) / 2
        
        # padding mask
        if mask is not None:
            batch_size = inputs.size()[0]
            seq_len = inputs.size()[1]
            pad_mask = mask.unsqueeze(1).unsqueeze(1).expand(batch_size, self.output_size, seq_len, seq_len)
            logits = logits*pad_mask - (1-pad_mask)*1e12

        # tril mask
        if self.tril_mask:
            mask = torch.tril(torch.ones_like(logits), -1) 
            logits = logits - mask * 1e12

        return logits


import torch
from typing import List
import numpy as np
import unicodedata
from torch.nn.utils.rnn import pad_sequence
import collections
import re


def match(s: str, text: str):
    """在一个文本中匹配特定字符串

    Args:
        s (str): 待匹配字符串
        text (str): 待匹配文本
    Returns
        starts(List): 匹配到的所有开始下标,如果没有则为空列表
    """
    if s == '':
        return []
    starts = []
    for i in range(0, len(text)):
        if text[i:i+len(s)] == s:
            starts.append(i)
    return starts
            
        
def truncate_sequences(maxlen, indices, *sequences):
    """截断总长度至不超过maxlen
    """
    sequences = [s for s in sequences if s]
    if not isinstance(indices, (list, tuple)):
        indices = [indices] * len(sequences)

    while True:
        lengths = [len(s) for s in sequences]
        if sum(lengths) > maxlen:
            i = np.argmax(lengths)
            sequences[i].pop(indices[i])
        else:
            return sequences

def lowercase_and_normalize(text):
    """转小写，并进行简单的标准化
    """
    text = text.lower()
    text = unicodedata.normalize('NFD', text)
    text = ''.join([ch for ch in text if unicodedata.category(ch) != 'Mn'])
    return text


def sequence_padding(inputs, length=None, value=0, seq_dims=1, mode='post'):
    """将序列padding到同一长度
    """
    if isinstance(inputs[0], (np.ndarray, list)):
        if length is None:
            length = np.max([np.shape(x)[:seq_dims] for x in inputs], axis=0)
        elif not hasattr(length, '__getitem__'):
            length = [length]

        slices = [np.s_[:length[i]] for i in range(seq_dims)]
        slices = tuple(slices) if len(slices) > 1 else slices[0]
        pad_width = [(0, 0) for _ in np.shape(inputs[0])]

        outputs = []
        for x in inputs:
            x = x[slices]
            for i in range(seq_dims):
                if mode == 'post':
                    pad_width[i] = (0, length[i] - np.shape(x)[i])
                elif mode == 'pre':
                    pad_width[i] = (length[i] - np.shape(x)[i], 0)
                else:
                    raise ValueError('"mode" argument must be "post" or "pre".')
            x = np.pad(x, pad_width, 'constant', constant_values=value)
            outputs.append(x)

        return np.array(outputs)
    
    elif isinstance(inputs[0], torch.Tensor):
        assert mode == 'post', '"mode" argument must be "post" when element is torch.Tensor'
        if length is not None:
            inputs = [i[:length] for i in inputs]
        return pad_sequence(inputs, padding_value=value, batch_first=True)
    else:
        raise ValueError('"input" argument must be tensor/list/ndarray.')
    
    
def load_vocab(vocab_file):
    """加载词典文件到dict"""
    vocab = collections.OrderedDict()
    index = 0
    with open(vocab_file, "r", encoding="utf-8") as reader:
        while True:
            token = reader.readline()
            if not token:
                break
            token = token.strip()
            vocab[token] = index
            index += 1
    return vocab


FH_SPACE = FHS = ((u"　", u" "),)
FH_NUM = FHN = (
    (u"０", u"0"), (u"１", u"1"), (u"２", u"2"), (u"３", u"3"), (u"４", u"4"),
    (u"５", u"5"), (u"６", u"6"), (u"７", u"7"), (u"８", u"8"), (u"９", u"9"),
)
FH_ALPHA = FHA = (
    (u"ａ", u"a"), (u"ｂ", u"b"), (u"ｃ", u"c"), (u"ｄ", u"d"), (u"ｅ", u"e"),
    (u"ｆ", u"f"), (u"ｇ", u"g"), (u"ｈ", u"h"), (u"ｉ", u"i"), (u"ｊ", u"j"),
    (u"ｋ", u"k"), (u"ｌ", u"l"), (u"ｍ", u"m"), (u"ｎ", u"n"), (u"ｏ", u"o"),
    (u"ｐ", u"p"), (u"ｑ", u"q"), (u"ｒ", u"r"), (u"ｓ", u"s"), (u"ｔ", u"t"),
    (u"ｕ", u"u"), (u"ｖ", u"v"), (u"ｗ", u"w"), (u"ｘ", u"x"), (u"ｙ", u"y"), (u"ｚ", u"z"),
    (u"Ａ", u"A"), (u"Ｂ", u"B"), (u"Ｃ", u"C"), (u"Ｄ", u"D"), (u"Ｅ", u"E"),
    (u"Ｆ", u"F"), (u"Ｇ", u"G"), (u"Ｈ", u"H"), (u"Ｉ", u"I"), (u"Ｊ", u"J"),
    (u"Ｋ", u"K"), (u"Ｌ", u"L"), (u"Ｍ", u"M"), (u"Ｎ", u"N"), (u"Ｏ", u"O"),
    (u"Ｐ", u"P"), (u"Ｑ", u"Q"), (u"Ｒ", u"R"), (u"Ｓ", u"S"), (u"Ｔ", u"T"),
    (u"Ｕ", u"U"), (u"Ｖ", u"V"), (u"Ｗ", u"W"), (u"Ｘ", u"X"), (u"Ｙ", u"Y"), (u"Ｚ", u"Z"),
)
FH_PUNCTUATION = FHP = (
    (u"．", u"."), (u"，", u","), (u"！", u"!"), (u"？", u"?"), (u"”", u'"'),
    (u"’", u"'"), (u"‘", u"`"), (u"＠", u"@"), (u"＿", u"_"), (u"：", u":"),
    (u"；", u";"), (u"＃", u"#"), (u"＄", u"$"), (u"％", u"%"), (u"＆", u"&"),
    (u"（", u"("), (u"）", u")"), (u"‐", u"-"), (u"＝", u"="), (u"＊", u"*"),
    (u"＋", u"+"), (u"－", u"-"), (u"／", u"/"), (u"＜", u"<"), (u"＞", u">"),
    (u"［", u"["), (u"￥", u"\\"), (u"］", u"]"), (u"＾", u"^"), (u"｛", u"{"),
    (u"｜", u"|"), (u"｝", u"}"), (u"～", u"~"), (u'“', u'"'), (u'。', u'.'),
)
FH_ASCII = HAC = lambda: ((fr, to) for m in (FH_ALPHA, FH_NUM, FH_PUNCTUATION) for fr, to in m)
HF_SPACE = HFS = ((u" ", u"　"),)
HF_NUM = HFN = lambda: ((h, z) for z, h in FH_NUM)
HF_ALPHA = HFA = lambda: ((h, z) for z, h in FH_ALPHA)
HF_PUNCTUATION = HFP = lambda: ((h, z) for z, h in FH_PUNCTUATION)
HF_ASCII = ZAC = lambda: ((h, z) for z, h in FH_ASCII())


def convert_FH(text, *maps, **ops):
    """ 全角/半角转换
    args:
        text: unicode string need to convert
        maps: conversion maps
        skip: skip out of character. In a tuple or string
        return: converted unicode string
    """
    if "skip" in ops:
        skip = ops["skip"]
        if isinstance(skip, str):
            skip = tuple(skip)
        def replace(text, fr, to):
            return text if fr in skip else text.replace(fr, to)
    else:
        def replace(text, fr, to):
            return text.replace(fr, to)
    for m in maps:
        if callable(m):
            m = m()
        elif isinstance(m, dict):
            m = m.items()
        for fr, to in m:
            text = replace(text, fr, to)

    return text

def fine_grade_tokenize(raw_text:str, tokenizer, convert_fh:bool =False) -> List[str]:
    """
    - 全角转半角
    - 序列标注任务 BERT 分词器可能会导致标注偏移，
    - 用 char-level 来 tokenize

    """
    tokens = []
    # 进行全角转半角预处理
    if convert_fh:
        raw_text = convert_FH(raw_text, FH_ASCII)  

    for _ch in raw_text:
        if _ch in [' ', '\t', '\n']:
            tokens.append('[BLANK]')
        else:
            if not len(tokenizer.tokenize(_ch)):
                tokens.append('[INV]')
            else:
                tokens.append(_ch)
    return tokens


_SEPARATOR = r'@'
_RE_SENTENCE = re.compile(r'(\S.+?[.!?])(?=\s+|$)|(\S.+?)(?=[\n]|$)', re.UNICODE)
_AB_SENIOR = re.compile(r'([A-Z][a-z]{1,2}\.)\s(\w)', re.UNICODE)
_AB_ACRONYM = re.compile(r'(\.[a-zA-Z]\.)\s(\w)', re.UNICODE)
_UNDO_AB_SENIOR = re.compile(r'([A-Z][a-z]{1,2}\.)' + _SEPARATOR + r'(\w)', re.UNICODE)
_UNDO_AB_ACRONYM = re.compile(r'(\.[a-zA-Z]\.)' + _SEPARATOR + r'(\w)', re.UNICODE)


def _replace_with_separator(text, separator, regexs):
    replacement = r"\1" + separator + r"\2"
    result = text
    for regex in regexs:
        result = regex.sub(replacement, result)
    return result

 
def split_sentence(text: str, best: bool =True):
    """将文本切分为句子

    Args:
        text (str): 原始文本
        best (bool, optional): . Defaults to True.
    """
    text = re.sub(r'([。！？?])([^”’])', r"\1\n\2", text)
    text = re.sub(r'(\.{6})([^”’])', r"\1\n\2", text)
    text = re.sub(r'(…{2})([^”’])', r"\1\n\2", text)
    text = re.sub(r'([。！？?][”’])([^，。！？?])', r'\1\n\2', text)
    for chunk in text.split("\n"):
        if not chunk:
            continue
        if not best:
            yield chunk
            continue
        processed = _replace_with_separator(chunk, _SEPARATOR, [_AB_SENIOR, _AB_ACRONYM])
        sents = list(_RE_SENTENCE.finditer(processed))
        if not sents:
            yield chunk
            continue
        for sentence in sents:
            sentence = _replace_with_separator(sentence.group(), r" ", [_UNDO_AB_SENIOR, _UNDO_AB_ACRONYM])
            yield sentence


from .gplinker import GPLinkerForRelationExtraction
from .onerel import OneRelForRelationExtraction
from .casrel import CasRelForRelationExtraction
from .biaffine import BLinkerForEntityRelationExtraction

__all__ = ['GPLinkerForRelationExtraction', 
           'OneRelForRelationExtraction',
           'CasRelForRelationExtraction',
           'BLinkerForEntityRelationExtraction']


import torch
from torchmetrics import F1Score
from typing import List, Tuple
from ...layers import SimpleDense, MultiDropout
from ...utils.make_model import PLMBaseModel


class BertForTextClassification(PLMBaseModel):
    '''
    文本分类模型
    '''
    def __init__(self, 
                 lr: float = 3e-5,
                 hidden_size: int = 256,
                 scheduler: str = 'linear_warmup',
                 weight_decay: float = 0.01,
                 **kwargs):
        super().__init__()  

        # 模型架构
        self.bert = self.get_plm_architecture(add_pooler_layer=True)
        self.classifier = SimpleDense(input_size=self.bert.config.hidden_size, 
                                      hidden_size=hidden_size, 
                                      output_size=len(self.hparams.id2label))
        self.dropout = MultiDropout()
        
        # 损失函数
        self.criterion = torch.nn.CrossEntropyLoss()

        # 评价指标
        self.train_f1 = F1Score(num_classes=len(self.hparams.id2label), average='macro', task='multiclass')
        self.val_f1= F1Score(num_classes=len(self.hparams.id2label), average='macro', task='multiclass')
        self.test_f1 = F1Score(num_classes=len(self.hparams.id2label), average='macro', task='multiclass')
        
    def setup(self, stage: str):
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.bert_transform)


    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
        x = x.pooler_output
        x = self.dropout(x)
        logits = self.classifier(x)  # (batch_size, output_size)
        return logits
    
        
    def shared_step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        label_ids = batch['label_ids']
        logits = self(input_ids, token_type_ids, attention_mask)
        loss = self.criterion(logits, label_ids)
        pred_ids = torch.argmax(logits, dim=-1)
        return loss, pred_ids, label_ids

    def training_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.train_f1(pred_ids, label_ids)
        self.log('train/f1', self.train_f1, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    def validation_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.val_f1(pred_ids, label_ids)
        self.log('val/f1', self.val_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def test_step(self, batch, batch_idx):
        loss, pred_ids, label_ids = self.shared_step(batch)
        self.test_f1(pred_ids, label_ids)
        self.log('test/f1', self.test_f1, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    
    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def predict(self, text: str, device: str='cpu') -> List[Tuple[str, int]]:
        device = torch.device(device)
        inputs = self.tokenizer(text,
                                padding='max_length',
                                max_length=self.hparams.max_length,
                                return_tensors='pt',
                                truncation=True)
        self.to(device)
        inputs.to(device)
        self.eval()
        with torch.no_grad():
            logits = self(**inputs)
            scores = torch.nn.functional.softmax(logits, dim=-1).tolist()
            cats = {}
            for i, v in enumerate(scores[0]):   # scores : [[0.1, 0.2, 0.3, 0.4]]
                cats[self.hparams.id2label[i]] = v
        return sorted(cats.items(), key=lambda x: x[1], reverse=True)


import os
from .utils import get_logger
from typing import Union, List
from torch.utils.data import DataLoader, BatchSampler, RandomSampler
from datasets import load_from_disk, load_dataset, DatasetDict
from transformers import AutoConfig, AutoTokenizer, AutoModel, PreTrainedTokenizerFast
from functools import lru_cache
from pathlib import Path
import numpy as np
from lightning.pytorch import LightningDataModule
from huggingface_hub import snapshot_download


log = get_logger()


def sequence_padding(inputs, length=None, value=0, seq_dims=1, mode='post'):
    """Numpy函数，将序列padding到同一长度
    """
    if length is None:
        length = np.max([np.shape(x)[:seq_dims] for x in inputs], axis=0)
    elif not hasattr(length, '__getitem__'):
        length = [length]

    slices = [np.s_[:length[i]] for i in range(seq_dims)]
    slices = tuple(slices) if len(slices) > 1 else slices[0]
    pad_width = [(0, 0) for _ in np.shape(inputs[0])]

    outputs = []
    for x in inputs:
        x = x[slices]
        for i in range(seq_dims):
            if mode == 'post':
                pad_width[i] = (0, length[i] - np.shape(x)[i])
            elif mode == 'pre':
                pad_width[i] = (length[i] - np.shape(x)[i], 0)
            else:
                raise ValueError('"mode" argument must be "post" or "pre".')
        x = np.pad(x, pad_width, 'constant', constant_values=value)
        outputs.append(x)
    return np.array(outputs)


def char_idx_to_token(char_idx, offset_mapping):
    """
    将char级别的idx 转换为token级别的idx
    例如:
    text = '我是中国人'
    tokens = ['我', '是', '中国', '人']
    '国'的字符下标为3, token级别的下标为2
    """
    for index, span in enumerate(offset_mapping):
        if span[0] <= char_idx < span[1]:
            return index
    return -1
        
        
def align_char_span(char_span_offset: tuple, 
                    token_offset_mapping, 
                    special_offset=(0,0),
                    add_special: bool=False) -> tuple:
    '''对齐字符级别的span标签与bert的子词切分的标签
    参数:
    - char_span_offsets: 字符级别的文本片段下标,例如(0,1)
    - token_offset_mapping: 词符与字符下标映射例如[(0,1), (1,2), (2,3) ...]
    输出
    token_span_offset: 词符级别的文本片段下标,例如(0,2)
    '''
    token_span_offset = ()
    bias = 0
    for i, offset in enumerate(token_offset_mapping):
        if offset == special_offset:
            bias += 1
            continue
        if offset != special_offset:
            if offset[0] == char_span_offset[0]:
                start = i
                if offset[1] == char_span_offset[1]:
                    end = i+1
                    break
                else: 
                    continue
            if offset[1] == char_span_offset[1]:
                end = i+1
                break
    try:
        if add_special:
            token_span_offset = (start, end)
        else:
            token_span_offset = (start-bias, end-bias)     
    except:
        log.warning('align offset failed')
    return token_span_offset

def align_char_span_text_b(char_span_offset: tuple, 
                            token_offset_mapping, 
                            special_offset=(0,0),
                            add_pre: bool=True) -> tuple:
    """对齐文本b的token下标, 适用于bert text pair形式

    Args:
        char_span_offset (tuple): 文本中的char级别的offset
        token_offset_mapping (_type_): 映射字典
        special_offset (tuple, optional): 特殊token offset. Defaults to (0,0).
        add_pre (bool, optional): 是否加上之前的token. Defaults to True.

    Returns:
        tuple: token级别的offset
    """
    token_span_offset = ()
    bias = 0
    for i, offset in enumerate(token_offset_mapping):
        if offset == special_offset and i ==0:
            continue
        if offset == special_offset:
            bias = i+1
        if offset != special_offset and bias>0:
            if offset[0] == char_span_offset[0]:
                start = i
                if offset[1] == char_span_offset[1]:
                    end = i+1
                    break
                else: 
                    continue
            if offset[1] == char_span_offset[1]:
                end = i+1
                break
    try:
        if add_pre:
            token_span_offset = (start, end)
        else:
            token_span_offset = (start-bias, end-bias)     
    except:
        log.warning(f'align {char_span_offset} failed')
    return token_span_offset


def prepare_dataset(dataset_name: str, dataset_dir) -> None:
    path = Path(dataset_dir, dataset_name)
    if path.exists():
        pass
    else:
        log.info('cannot found dataset in {}.'.format(path))
        try:
            log.info(f'download dataset {dataset_name} from huffingface')
            snapshot_download(dataset_name, 
                              repo_type='dataset',
                              endpoint="https://hf-mirror.com")
            dataset = load_dataset(dataset_name)
            dataset.save_to_disk(path)
            log.info(f'download dataset succeed')
        except Exception as e:
            log.error('download dataset failed')
            raise(e)
            
            
def prepare_plm(plm_name: str, plm_dir: str) -> None:
    path = Path(plm_dir, plm_name)
    if path.exists():
        pass
    else : 
        log.info('cannot found plm in {}'.format(path))
        try:
            log.info(f'download plm {plm_name} from huffingface')
            model = AutoModel.from_pretrained(plm_name)
            tokenizer = AutoTokenizer.from_pretrained(plm_name)
            model.save_pretrained(path)
            tokenizer.save_pretrained(path)
            log.info(f'download plm succeed')
        except:
            log.error('download plm failed')
        


def prepare_from_huffingface(dataset: str, 
                             dataset_dir: str,
                             plm: str,
                             plm_dir: str) -> None:
        '''
        下载数据集和预训练模型这个方法只会在一个GPU上执行一次.
        '''
        prepare_dataset(dataset_name=dataset, dataset_dir=dataset_dir)
        prepare_plm(plm_name=plm, plm_dir=plm_dir)


class PLMBaseDataModule(LightningDataModule):
    """数据模块的基类,子类需要完成setup方法,子类初始化的时候至少包含dataset,plm,batch_size参数,
    内置功能:
    - 自动保存超参数
    - 不同策略自动获取最大文本长度,超过512则取512
    - 下载数据集和预训练模型
    - 自动读取tokenizer
    - 自动读取数据集
    - 自动设置dataloader,数据集需要切分为train,validation,test
    """
    def __init__(self,
                 auto_length: Union[str, int] = 'max',
                 plm_dir: str = 'plms',
                 plm_max_length: int = 512,
                 dataset_dir: str = 'datasets',
                 num_workers: int = 4,
                 pin_memory: bool = False,
                 shuffle_train: bool = False,
                 shuffle_val: bool = False,
                 shuffle_test: bool = False):
        super().__init__()
        self.save_hyperparameters()
        self.transforms = {}
    
    
    def prepare_data(self) -> None:
        prepare_from_huffingface(dataset=self.hparams.dataset, 
                                 dataset_dir=self.hparams.dataset_dir,
                                 plm=self.hparams.plm,
                                 plm_dir=self.hparams.plm_dir)
    
    @property
    @lru_cache()
    def dataset(self) -> DatasetDict:
        dataset_path = Path(self.hparams.dataset_dir, self.hparams.dataset)
        if dataset_path.exists():  
            dsd = load_from_disk(dataset_path)
        else:
            dsd = load_dataset(self.hparams.dataset)
        return dsd
            
            
    @lru_cache()
    def get_trf_config(self):
        plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
        config = AutoConfig.from_pretrained(plm_path)
        config = config.to_dict()
        return config
    
    
    @property
    @lru_cache()
    def tokenizer(self) -> PreTrainedTokenizerFast:
        plm_path = os.path.join(self.hparams.plm_dir, self.hparams.plm)
        return AutoTokenizer.from_pretrained(plm_path)
    
    
    def get_vocab(self):
        return dict(sorted(self.tokenizer.vocab.items(), key=lambda x: x[1]))


    def get_available_transforms(self):
        return self.transforms.keys()
    
        
    @lru_cache()
    def get_max_length(self):
        """根据auto_length参数自动获取最大token的长度
        
        Returns:
            int: 最大token长度
        """
        length = self.train_df.text.map(lambda x: len(self.tokenizer.tokenize(x)))
        if self.hparams.auto_length == 'max':
            max_length = length.max()
        if self.hparams.auto_length == 'mean':
            max_length = int(length.mean())
        if type(self.hparams.auto_length) == int:
            assert self.hparams.auto_length >0, 'max_length length  must > 0'
            max_length = self.hparams.auto_length
        return max_length
    
    
    def get_batch_max_length(self, batch_text: List[str]) -> int:
        """获取一个batch的最大token长度,不会大于预训练模型的最大输入长度,一般用于dataset的transform中

        Args:
            batch_text (List[str]): 一个批次的文本

        Returns:
            int: 最大文本长度
        """
        max_length = max([len(self.tokenizer.encode(t)) for t in batch_text]) # 获取最大token序列的长度
        max_length = min([self.hparams.plm_max_length, max_length])
        return max_length
    
    
    @property
    @lru_cache()
    def train_df(self):
        return self.dataset['train'].to_pandas()
    
    @property
    @lru_cache()
    def val_df(self):
        return self.dataset['validation'].to_pandas()
    
    @property
    @lru_cache()
    def test_df(self):
        return self.dataset['test'].to_pandas()
    
    def train_dataloader(self):
        return DataLoader(dataset= self.dataset['train'], 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_train,
                          batch_size=None,
                          sampler=BatchSampler(RandomSampler(self.dataset['train']), batch_size=self.hparams.batch_size, drop_last=False))
    
    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_val,
                          sampler=BatchSampler(RandomSampler(self.dataset['validation']), batch_size=self.hparams.batch_size, drop_last=False))

    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_test,
                          sampler=BatchSampler(RandomSampler(self.dataset['test']), batch_size=self.hparams.batch_size, drop_last=False))


class BaseDataModule(LightningDataModule):
    def __init__(self,
                 num_workers: int = 0,
                 pin_memory: bool = True,
                 shuffle_train: bool = False,
                 shuffle_val: bool = False,
                 shuffle_test: bool = False,
                 drop_last: bool = False):
        super().__init__()
        self.save_hyperparameters()
        assert 'batch_size' in self.hparams and 'dataset_path' in self.hparams and 'tokenizer_path' in self.hparams, '子类至少需要传入dataset_path, tokenizer_path, batch_size参数'
        
    @property
    @lru_cache()
    def dataset(self):
        path = Path(self.hparams.dataset_path)
        if path.exists():
            return load_from_disk(path)
        else:
            return load_dataset(path=self.hparams.dataset_path)
        
    @property
    @lru_cache()
    def tokenizer(self):
        return AutoTokenizer.from_pretrained(self.hparams.tokenizer_path)
        
    @property
    @lru_cache()
    def train_df(self):
        return self.dataset['train'].to_pandas()
    
    @property
    @lru_cache()
    def val_df(self):
        return self.dataset['validation'].to_pandas()
    
    @property
    @lru_cache()
    def test_df(self):
        return self.dataset['test'].to_pandas()
    
    def train_dataloader(self):
        return DataLoader(dataset= self.dataset['train'], 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_train,
                          batch_size=None,
                          sampler=BatchSampler(RandomSampler(self.dataset['train']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last))
    
    def val_dataloader(self):
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_val,
                          sampler=BatchSampler(RandomSampler(self.dataset['validation']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last))

    def test_dataloader(self):
        return DataLoader(dataset=self.dataset['test'], 
                          batch_size=None, 
                          num_workers=self.hparams.num_workers, 
                          pin_memory=self.hparams.pin_memory,
                          shuffle=self.hparams.shuffle_test,
                          sampler=BatchSampler(RandomSampler(self.dataset['test']), batch_size=self.hparams.batch_size, drop_last=self.hparams.drop_last))


import torch
from ...metrics.span import SpanF1
from ...utils.make_model import PLMBaseModel
from ...layers import MultiLabelCategoricalCrossEntropy, EfficientGlobalPointer, MultiDropout
from ...tricks.adversarial_training import adversical_tricks
from ...data.doc import Entity
from typing import Optional


class GlobalPointerForEntityExtraction(PLMBaseModel):
    """基于EfficientGlobalPointer的实体识别模型,可以统一处理嵌套和非嵌套模型
    
    参考:
        - https://kexue.fm/archives/8373
        
    参数:
        - lr (float): 学习率
        - hidden_size (int): globalpointer中间层维度
        - scheduler (str): 学习率变化器,[harmonic_epoch, linear_warmup_step, cosine_warmup_step]之一
        - weight_decay (float): 权重衰减
        - adv Optinal[str]: 对抗训练方式, fgm, pgd之一
        - threshold (float): 阈值
        - add_rope (bool): 是否添加RoPE位置矩阵
        - **kwargs datamodule的hparams
    """
    def __init__(self,
                 lr: float =3e-5,
                 hidden_size: int = 64,
                 scheduler: str = 'linear_warmup',
                 weight_decay: float =0.01,
                 adv: Optional[str] = None,
                 threshold: float = 0.0,
                 **kwargs) : 
        super().__init__()
        ## 手动optimizer 可参考https://pytorch-lightning.readthedocs.io/en/stable/common/optimizers.html#manual-optimization
        self.automatic_optimization = False    


        self.plm = self.get_plm_architecture()
        
        self.classifier = EfficientGlobalPointer(input_size=self.plm.config.hidden_size, 
                                                hidden_size=hidden_size,
                                                output_size=len(self.hparams.id2ent),
                                                add_rope=True)

        self.dropout = MultiDropout()
        self.criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = SpanF1()
        self.val_metric = SpanF1()
        self.test_metric = SpanF1()
        
    def setup(self, stage: Optional[str] = None) -> None:
        self.trainer.datamodule.dataset.set_transform(self.trainer.datamodule.tp_transform)


    def forward(self, input_ids, attention_mask=None):
        x = self.plm(input_ids=input_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        logits = self.classifier(x, mask=attention_mask)
        return logits


    def on_train_start(self) -> None:
        if self.hparams.adv :
            self.adv = adversical_tricks.get(self.hparams.adv)(self.bert)


    def shared_step(self, batch):
        span_ids = batch['tag_ids']
        logits = self(input_ids=batch['input_ids'], attention_mask=batch['attention_mask'])
        pred = logits.ge(self.hparams.threshold).float()
        batch_size, ent_type_size = logits.shape[:2]
        y_true = span_ids.reshape(batch_size*ent_type_size, -1)
        y_pred = logits.reshape(batch_size*ent_type_size, -1)
        loss = self.criterion(y_pred, y_true)
        return loss, pred, span_ids


    def training_step(self, batch, batch_idx):
        optimizer = self.optimizers()
        optimizer.zero_grad()
        scheduler = self.lr_schedulers()
        loss, pred, true = self.shared_step(batch)
        self.manual_backward(loss)
        if self.hparams.adv == 'fgm':
            self.adv.attack()
            loss_adv,  _,  _ = self.shared_step(batch)
            self.manual_backward(loss_adv)
            self.adv.restore()
            self.log_dict({'train_loss': loss, 'adv_loss': loss_adv}, prog_bar=True)
        elif self.hparams.adv == "pgd":
            self.adv.backup_grad()
            K=3
            for t in range(K):
                self.adv.attack(is_first_attack=(t==0)) # 在embedding上添加对抗扰动, first attack时备份param.data
                if t != K-1:
                    self.zero_grad()
                else:
                    self.adv.restore_grad()
                loss_adv, _, _ = self.shared_step(batch)
                self.manual_backward(loss_adv) # 反向传播，并在正常的grad基础上，累加对抗训练的梯度
            self.adv.restore()
            self.log_dict({'train_loss': loss, 'adv_loss': loss_adv}, prog_bar=True)
        optimizer.step()
        if self.hparams.scheduler in ['harmonic_epoch']:
            if self.trainer.is_last_batch:
                scheduler.step()
        else: scheduler.step()
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, on_epoch=True, prog_bar=True)
        self.log_dict({'train_loss': loss}, prog_bar=True)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def test_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.test_metric(pred, true)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def predict(self, text: str, device: str='cpu'):
        threshold = self.hparams.threshold
        inputs = self.tokenizer(text,
                                max_length=self.hparams.plm_max_length,
                                truncation=True,
                                return_token_type_ids=False,
                                return_tensors='pt')
        inputs.to(device)
        logits = self(**inputs)
        spans_ls = torch.nonzero(logits>threshold).tolist()
        ents = []
        for span in spans_ls :
            start_token = span[2]
            start_offset = inputs.token_to_chars(start_token)
            start = start_offset[0]
            end_token = span[3]
            end_offset = inputs.token_to_chars(end_token)
            end = end_offset[-1]
            span_text = text[start:end]
            label = self.hparams.id2ent[span[1]]
            ent = Entity(text=span_text, indices=[i for i in range(start, end)], label=label)
            ents.append(ent)
        return ents


from pydantic import BaseModel, constr, validator, validate_arguments
from typing import List, Union, Optional
from .dataset import Dataset
import pandas as pd
from pathlib import Path
import srsly


class Couplet(BaseModel):
    """一副对联
    Args:
        left: 上联
        right: 下联
        top: 横批
    """
    left: constr(strip_whitespace=True, strict=True, min_length=1)
    right: constr(strip_whitespace=True, strict=True, min_length=1)
    top: constr(strip_whitespace=True, strict=True, min_length=0) = None
    
    @validator('right')
    def validate_right(cls, v: str, values) -> str:
        left = values['left']
        right = v.strip()
        assert len(right) == len(left), '上联与下联字数不同'
        return right

class CoupletBin:
    """保存对联
    """
    def __init__(self, couplets: Union[List[Couplet], Path] = []) -> None:
        super().__init__()
        if type(couplets) == str:
            self._couplets = self._get_docs_from_jsonl(couplets)
        elif couplets is None:
            self._couplets = []
        else:
            self._couplets = couplets
    
    def __getitem__(self, i):
        return self._couplets[i]
    
    def __len__(self):
        return len(self._couplets)
    
    def __repr__(self) -> str:
        return f"{len(self._couplets)} couplets"
    
    def __str__(self) -> str:
        return f"{len(self._couplets)} couplets"
    
    def __add__(self, other: Union["CoupletBin", List[Couplet]]) -> "CoupletBin":
        if isinstance(other, list):
            return CoupletBin(self._couplets + other)
        if isinstance(other, CoupletBin):
            return CoupletBin(self._couplets + other._couplets)
    
    @validate_arguments
    def append(self, couplet: Couplet):
        self._couplets.append(couplet)
        return self
        
    @validate_arguments
    def add(self, couplet: Couplet):
        if couplet not in self._couplets:
            self._couplets.append(couplet)
        return self
        
    def save_to_disk(self, file_path: Path):
        """将数据以jsonl的格式保存到硬盘
        参数:
        - file_path (Path): 数据保存地址,例如./test.jsonl
        """
        path = Path(file_path)
        srsly.write_jsonl(path=path, lines=[couplet.dict() for couplet in self._couplets])
    
    def _get_docs_from_jsonl(self, file_path: Path) -> List[Couplet]:
        path = Path(file_path)
        lines = srsly.read_jsonl(path=path)
        couplets = []     
        for l in lines:
            couplets.append(Couplet(**l))
        return couplets
        
    def to_dataset(self, include: Optional[List] = None) -> Dataset:
        """转换数据集,None的数据自动去除
        参数:
        - include (List): 包含的字段名称,默认None
        """
        if include:
            data = [c.dict(include=set(include), exclude_none=True) for c in self._couplets]
        else:
            data = [c.dict(exclude_none=True) for c in self._couplets]
        return Dataset.from_pandas(pd.DataFrame.from_dict(data))


from ...utils.make_model import PLMBaseModel, align_token_span
from ...layers.classifier import EfficientGlobalPointer
from ...layers.loss import SparseMultiLabelCrossEntropy
from ...layers.dropout import MultiDropout
from ...metrics.event import EventF1, Event, Entity
from ...metrics.span import SpanF1
import torch
from itertools import groupby
from typing import Any, Union, Optional, List, Tuple
import numpy as np


class DedupList(list):
    """定义去重的list
    """
    def append(self, x):
        if x not in self:
            super(DedupList, self).append(x)


def clique_search(argus, links):
    """搜索每个节点所属的完全子图作为独立事件
    搜索思路：找出不相邻的节点，然后分别构建它们的邻集，递归处理。
    """
    def neighbors(host, argus, links):
        """构建邻集（host节点与其所有邻居的集合）
        """
        results = [host]
        for argu in argus:
            if host[2:] + argu[2:] in links:
                results.append(argu)
        return list(sorted(results))
    
    Argus = DedupList()
    for i1, (_, _, h1, t1) in enumerate(argus):
        for i2, (_, _, h2, t2) in enumerate(argus):
            if i2 > i1:
                if (h1, t1, h2, t2) not in links:
                    Argus.append(neighbors(argus[i1], argus, links))
                    Argus.append(neighbors(argus[i2], argus, links))
    if Argus:
        results = DedupList()
        for A in Argus:
            for a in clique_search(A, links):
                results.append(a)
        return results
    else:
        return [list(sorted(argus))]

class GPLinkerForEventExtraction(PLMBaseModel):
    def __init__(self, 
                 lr: float = 3e-5,
                 hidden_size: int = 64,
                 weight_decay: float = 0.01,
                 threshold: float = 0.0,
                 scheduler: str = 'linear_warmup',
                 **kwargs: Any) -> None:
        super().__init__()
        
        self.plm = self.get_plm_architecture()

        self.role_classifier = EfficientGlobalPointer(input_size=self.plm.config.hidden_size,
                                                      hidden_size=hidden_size,
                                                      output_size=len(self.hparams.id2combined))
        
        self.head_classifier = EfficientGlobalPointer(input_size=self.plm.config.hidden_size,
                                                      hidden_size=hidden_size,
                                                      output_size=1,
                                                      add_rope=False)
        
        self.tail_classifier = EfficientGlobalPointer(input_size=self.plm.config.hidden_size,
                                                      hidden_size=hidden_size,
                                                      output_size=1,
                                                      add_rope=False)
        
        
        self.dropout = MultiDropout()
        
        self.role_criterion = SparseMultiLabelCrossEntropy()
        self.head_criterion = SparseMultiLabelCrossEntropy()
        self.tail_criterion = SparseMultiLabelCrossEntropy()
                 
        self.train_role_metric = SpanF1()
        self.train_head_metric = SpanF1()
        self.train_tail_metric = SpanF1()
        self.val_role_metric = SpanF1()
        self.val_head_metric = SpanF1()
        self.val_tail_metric = SpanF1()
        # 以事件级别的f1最为最终指标
        self.val_event_metric = EventF1()
        
        
    def setup(self, stage: str) -> None:
        self.trainer.datamodule.dataset['train'].set_transform(self.trainer.datamodule.sparse_combined_transform)
        self.trainer.datamodule.dataset['validation'].set_transform(self.trainer.datamodule.combined_transform)
        
        
    def forward(self, input_ids, attention_mask):
        x = self.plm(input_ids=input_ids,attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        role_logits = self.role_classifier(x, mask=attention_mask)
        head_logits = self.head_classifier(x, mask=attention_mask)
        tail_logits = self.tail_classifier(x, mask=attention_mask)
        return role_logits, head_logits, tail_logits


    def step(self, batch, is_train: bool=True):
        
        input_ids = batch['input_ids']
        attention_mask = batch['attention_mask']
        role_logits, head_logits, tail_logits = self(input_ids, attention_mask)
        role_true = batch['role_tags']
        head_true = batch['head_tags']
        tail_true = batch['tail_tags']
    
        if is_train:
            b,c,s,s = role_logits.shape
            role_logits = role_logits.reshape(b,c,s*s)
            role_true = role_true[..., 0] * s + role_true[..., 1]
            role_loss = self.role_criterion(role_logits, role_true)
        
            b,c,s,s = head_logits.shape
            head_logits = head_logits.reshape(b,c,s*s)
            head_true = head_true[..., 0] * s + head_true[..., 1]
            head_loss = self.head_criterion(head_logits, head_true)
            
            b,c,s,s = tail_logits.shape
            tail_logits = tail_logits.reshape(b,c,s*s)
            tail_true = tail_true[..., 0] * s + tail_true[..., 1]
            tail_loss = self.tail_criterion(tail_logits, tail_true)
            self.log('train/role_loss', role_loss, prog_bar=True, on_step=True)
            self.log('train/head_loss', head_loss, prog_bar=True, on_step=True)
            self.log('train/tail_loss', tail_loss, prog_bar=True, on_step=True)
            
            loss = sum([role_loss, head_loss, tail_loss]) / 3
        else:
            b, c, s, s = role_logits.shape
            role_loss = self.role_criterion(role_logits.reshape(b, -1, s*s), role_true.reshape(b, -1, s*s))
            head_loss = self.head_criterion(head_logits.reshape(b, -1, s*s), head_true.reshape(b, -1, s*s))
            tail_loss = self.tail_criterion(tail_logits.reshape(b, -1, s*s), tail_true.reshape(b, -1, s*s))
            loss = sum([role_loss, head_loss, tail_loss]) / 3
            
        
        return loss, role_logits, head_logits, tail_logits, role_true, head_true, tail_true


    def training_step(self, batch, batch_idx):
        loss, role_logits, head_logits, tail_logits, role_true, head_true, tail_true = self.step(batch)
        return {'loss': loss}


    def validation_step(self, batch, batch_idx):
        loss, role_logits, head_logits, tail_logits, role_true, head_true, tail_true = self.step(batch, is_train=False)
        batch_true_events = self.extract_events(role_true, head_true, tail_true)
        batch_pred_events = self.extract_events(role_logits, head_logits, tail_logits)
        self.val_event_metric(batch_pred_events, batch_true_events)
        self.log('val/f1', self.val_event_metric, on_step=False, on_epoch=True, prog_bar=True)
        
        role_pred = role_logits.gt(self.hparams.threshold).float()
        head_pred = head_logits.gt(self.hparams.threshold).float()
        tail_pred = tail_logits.gt(self.hparams.threshold).float()
        self.val_role_metric(role_pred, role_true)
        self.val_head_metric(head_pred, head_true)
        self.val_tail_metric(tail_pred, tail_true)
        self.log('val/role_f1', self.val_role_metric, on_epoch=True, prog_bar=True)
        self.log('val/head_f1', self.val_head_metric, on_epoch=True, prog_bar=True)
        self.log('val/tail_f1', self.val_tail_metric, on_epoch=True, prog_bar=True)
        return {'val_loss': loss}


    def extract_events(self, role_logits: torch.Tensor, head_logits: torch.Tensor, tail_logits: torch.Tensor, threshold: Union[None, float]= None, mapping: Optional[List[Tuple]]= None):
        if threshold is None:
            threshold = self.hparams.threshold
        role_logits = role_logits.cpu().detach().numpy()
        head_logits = head_logits.cpu().detach().numpy()
        tail_logits = tail_logits.cpu().detach().numpy()
        assert len(head_logits) == len(tail_logits) == len(role_logits)
        batch_events = []
        for role_logit, head_logit, tail_logit in zip(role_logits, head_logits, tail_logits):
            roles = set()
            for l, h, t in zip(*np.where(role_logit > threshold)):
                roles.add(tuple(self.hparams.id2combined[l.item()]) + (h.item(), t.item()))
            links = set()
            for i1, (_, _, h1, t1) in enumerate(roles):
                for i2, (_, _, h2, t2) in enumerate(roles):
                    if i2 > i1:
                        # head_logits[i] 1,1,seq_len, seq_len
                        if head_logit[0, min(h1, h2), max(h1, h2)] > threshold:
                            if tail_logit[0, min(t1, t2), max(t1, t2)] > threshold:
                                links.add((h1, t1, h2, t2))
                                links.add((h2, t2, h1, t1))
            events = set()
            for e_label, sub_argus in groupby(sorted(roles), key=lambda x: x[0]):
                for event in clique_search(list(sub_argus), links):
                    role_ls = []
                    for argu in event:
                        role_label, start, end =argu[1], argu[2], argu[3]+1
                        if mapping is not None:
                            (start, end) = align_token_span((start, end), mapping)
                        role_ls.append(Entity(label=role_label, indices=[i for i in range(start, end)]))    
                                            
                    event = Event(label=e_label, args=role_ls)
                    events.add(event)
            batch_events.append(events)
        return batch_events

    
    def configure_optimizers(self)  :
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_params = [
            {'params': [p for n, p in self.named_parameters() if not any(nd in n for nd in no_decay)],
            'weight_decay': self.hparams.weight_decay, 'lr': self.hparams.lr},
            {'params': [p for n, p in self.named_parameters() if any(nd in n for nd in no_decay)],
            'weight_decay': 0.0, 'lr': self.hparams.lr}
        ]
        
        optimizer = torch.optim.AdamW(grouped_params)
        scheduler_config = self.get_scheduler_config(optimizer=optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]


    def predict(self, text: str, device: str = 'cpu', threshold: Optional[float] = None):
        if threshold is None:
            threshold = self.hparams.threshold
        inputs = self.tokenizer(text,
                                add_special_tokens=True,
                                max_length=self.hparams.plm_max_length,
                                padding=True,
                                truncation=True,
                                return_offsets_mapping=True,
                                return_tensors='pt',
                                return_token_type_ids=False)
        mapping = inputs.pop('offset_mapping')
        mapping = mapping[0].tolist()
        inputs.to(device)
        role_logits, head_logits, tail_logits = self(**inputs)
        events = self.extract_events(role_logits, head_logits, tail_logits, threshold=threshold, mapping=mapping)
        return [e for e in events[0]]


import requests


def augment_text_from_youdao_translator(text: str):
    """通过调用有道翻译接口增加文本数据,最多可以扩充三条文本

    Args:
        text (str): 原始文本

    Returns:
        List[str]: 增强的文本,不包含原始文本. 
    """
    results = []
    trans_types = [('ZH_CN2EN','EN2ZH_CN'),('ZH_CN2KR','KR2ZH_CN'),('ZH_CN2JA','JA2ZH_CN'),]
    for t in trans_types:
        try:
            data1 = { 'doctype': 'json', 'type': t[0],'i': text }
            r = requests.get("http://fanyi.youdao.com/translate",params=data1)
            r = r.json()
            tgt = r['translateResult'][0][0]['tgt']
            
            data2 = { 'doctype': 'json', 'type': t[1],'i': tgt}
            r = requests.get("http://fanyi.youdao.com/translate",params=data2)
            r = r.json()
            res = r['translateResult'][0][0]['tgt']
            if res not in results and res!=text:
                results.append(res)
        except:
            pass
    return results  


from ...utils.make_model import PLMBaseModel, align_token_span
from ...layers.dropout import MultiDropout
from ...metrics.triple import TripleF1, Triple
import torch.nn as nn
import torch


class OneRelClassifier(nn.Module):
    
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 tag_size: int,
                 dropout: float):
        """用一个矩阵表示两两token关系

        Args:
            input_size (int): 一般为bert的hidden_size
            output_size (int): 关系类别
            tag_size (int): 矩阵每个位置的分类类别
            dropout(float): dropout的比率
        """
        super().__init__()
        self.output_size = output_size
        self.tag_size =tag_size
        self.lin1 = nn.Linear(input_size*2, input_size*2)
        self.lin2 = nn.Linear(input_size*2, tag_size*output_size)
        self.activation = nn.ReLU()  
        self.dropout = MultiDropout()     
    def forward(self, x): 
        batch_size, seq_len, hidden_size = x.size()
        # head: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,1,1, 2,2,2, 3,3,3
        head = x.unsqueeze(2).expand(batch_size, seq_len, seq_len, hidden_size).reshape(batch_size, seq_len*seq_len, hidden_size)
        # tail: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,2,3, 1,2,3, 1,2,3
        tail = x.repeat(1, seq_len, 1)
        # 两两token的特征拼接在一起
        pairs = torch.cat([head, tail], dim=-1)
        pairs = self.lin1(pairs)
        pairs = self.dropout(pairs)
        pairs = self.activation(pairs)
        scores = self.lin2(pairs).reshape(batch_size, seq_len, seq_len,  self.output_size, self.tag_size)
        return scores



class OneRelForRelationExtraction(PLMBaseModel):
    def __init__(self,
                 lr: float,
                 scheduler: str,
                 dropout: float,
                 weight_decay: float,
                 **kwargs):
        super().__init__()
        
        self.plm = self.get_plm_architecture()
        self.dropout = MultiDropout()
        self.classifier = OneRelClassifier(input_size=self.plm.config.hidden_size, 
                                           output_size=len(self.hparams.label2id), 
                                           tag_size=len(self.hparams.tag2id),
                                           dropout=self.hparams.dropout)

        self.criterion = nn.CrossEntropyLoss(reduction='none')
        
        self.train_metric = TripleF1()
        self.val_metric = TripleF1()
        self.test_metric = TripleF1()

        
    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.plm(input_ids=input_ids, 
                     token_type_ids=token_type_ids, 
                     attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        x = self.classifier(x)
        return x
    
    def step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        target = batch['tag_ids']
        logits = self(input_ids, token_type_ids, attention_mask)
        loss = self.criterion(logits.permute(0, 4, 3, 1, 2), target)
        loss_mask = batch['loss_mask']
        loss = torch.sum(loss * loss_mask) / torch.sum(loss_mask) 
        tag_ids = logits.argmax(-1).permute(0, 3, 1, 2)
        return loss, tag_ids

    def training_step(self, batch, batch_idx):
        loss, tag_ids = self.step(batch)
        self.log('train/loss', loss)
        return {'loss':loss}
    
        
    def validation_step(self, batch, batch_idx):
        loss, tag_ids = self.step(batch)
        batch_triples_x = self.extract_triples(tag_ids)
        batch_triples_y = self.extract_triples(batch['tag_ids'])
        self.val_metric(batch_triples_x, batch_triples_y)
        self.log('val/f1', self.val_metric, on_epoch=True, on_step=False, prog_bar=True)


    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr*50, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr*50, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, name=self.hparams.scheduler)
        return [optimizer], [scheduler_config]

    def extract_triples(self, tag_ids):
        # 原作者解码实现
        batch_tag_ids = torch.chunk(tag_ids, tag_ids.shape[0])
        batch_triples = []
        tag2id = self.hparams.tag2id
        id2rel = self.hparams.id2label
        for i in range(len(batch_tag_ids)):
            tag_ids = batch_tag_ids[i].squeeze(0)
            rel_numbers, seq_lens, seq_lens = tag_ids.shape
            relations, heads, tails = torch.where(tag_ids > 0)
            triples = set()
            pair_numbers = len(relations)
            if pair_numbers > 0:
                for i in range(pair_numbers):
                    r_index = relations[i]
                    h_start_index = heads[i]
                    t_start_index = tails[i]
                    # 如果当前第一个标签为HB-TB
                    if tag_ids[r_index][h_start_index][t_start_index] == tag2id[
                        'HB-TB'] and i + 1 < pair_numbers:
                        # 如果下一个标签为HB-TE
                        t_end_index = tails[i + 1]
                        if tag_ids[r_index][h_start_index][t_end_index] == tag2id['HB-TE']:
                            # 那么就向下找
                            for h_end_index in range(h_start_index, seq_lens):
                                # 向下找到了结尾位置
                                if tag_ids[r_index][h_end_index][t_end_index] == tag2id['HE-TE']:

                                    sub_head, sub_tail = h_start_index.item(), h_end_index
                                    obj_head, obj_tail = t_start_index.item(), t_end_index.item()
                                    rel = id2rel[r_index.item()]
                                    triples.add(Triple(triple=(sub_head, sub_tail+1, rel, obj_head, obj_tail+1)))
                                    break
            batch_triples.append(triples)
        return batch_triples
        # 改进版本解码
        # for i in range(len(bath_tag_ids)):
        #     triples = set()
        #     if len(torch.nonzero(bath_tag_ids[i])) >0:
        #         logits = bath_tag_ids[i].squeeze(0)
        #         rel_num, seq_len, seq_len = logits.shape
        #         rel2tag = {}
        #         for rel_id, start, end in torch.nonzero(logits):
        #             rel_id = rel_id.item()
        #             start = start.item()
        #             end = end.item()
        #             tag = self.hparams.id2tag[logits[rel_id, start, end].item()]
        #             if rel_id not in rel2tag:
        #                 rel2tag[rel_id] = [] 
        #             rel2tag[rel_id].append((tag, start, end))
        #         for rel_id, tags in rel2tag.items():
        #             rel = self.hparams.id2label[rel_id]
        #             for i, tag in enumerate(tags):
        #                 if tag[0] == 'HB-TB': 
        #                     if i == len(tags)-1:  # 主体客体都是一个token的情况
        #                         triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tag[2]+1)))
        #                     if i<len(tags)-1 :
        #                         if tags[i+1][0] == 'HB-TB':  # 此时客体是一个token, 继续搜索
        #                             for j in range(tag[1], seq_len):
        #                                 tag_id = logits[rel_id, j, tag[2]]
        #                                 tag_label = self.hparams.id2tag[tag_id.item()]
        #                                 if tag_label == 'HE-TE': # 客体是一个token
        #                                     triples.add(Triple((tag[1], j+1, rel, tag[2], tag[2]+1)))
        #                                     break
        #                                 if tag_label == 'HB-TB': # 说明主体客体都是一个token
        #                                     triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tag[2]+1)))
        #                                     break
        #                                 if j == seq_len-1 and tag != 'HB-TB' and tag!= 'HE-TE':  #最后一个也没有搜到说明主体客体都是一个token
        #                                     triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tag[2]+1)))
        #                                     break
        #                         if tags[i+1][0] == 'HB-TE': # 此时客体不是一个token, 继续搜索主体
        #                             for j in range(tags[i+1][1], seq_len):
        #                                 tag_id = logits[rel_id, j, tags[i+1][2]]
        #                                 tag_label = self.hparams.id2tag[tag_id.item()]
        #                                 if tag_label == 'HE-TE': # 找到主体
        #                                     triples.add(Triple((tags[i+1][1], j+1, rel, tag[2], tags[i+1][2]+1)))
        #                                     break
        #                                 if tag_label == 'HB-TB': # 说明主体一个token
        #                                     triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tag[2]+1)))
        #                                     break
        #                                 if j == seq_len-1 and tag != 'HB-TB' and tag!= 'HE-TE':  #最后一个也没有搜到说明主体是一个token
        #                                     triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tag[2]+1)))
        #                                     break
        #                         if tags[i+1][0] == 'HE-TE' and tag[1] == tags[i+1][1]: # 此时主体是一个token
        #                             triples.add(Triple((tag[1], tag[1]+1, rel, tag[2], tags[i+1][2])))
        #     batch_triples.append(triples)  
        # return batch_triples
                                                           
    def predict(self, text, device: str):
        max_length = min(len(text), self.hparams.max_length)
        inputs = self.tokenizer(
                text, 
                padding='max_length',  
                max_length=max_length,
                return_tensors='pt',
                truncation=True)
        inputs.to(torch.device(device))
        logits = self(**inputs)
        tag_ids = logits.argmax(-1).permute(0, 3, 1, 2)
        triples = self.extract_triples(tag_ids)
        rels = []
        for batch_triples in triples:
            if len(batch_triples)>0:
                offset_mapping = self.tokenizer(text,
                                                max_length=max_length,
                                                padding='max_length',
                                                truncation=True,
                                                return_offsets_mapping=True)['offset_mapping']
                for triple in batch_triples:
                    sub = align_token_span((triple[0], triple[1]), offset_mapping)
                    obj = align_token_span((triple[3], triple[4]), offset_mapping)
                    rels.append((sub[0],sub[1],triple[2],obj[0],obj[1]))
        return rels


from ..utils.make_datamodule import PLMBaseDataModule, get_logger, char_idx_to_token, sequence_padding
from functools import lru_cache
import torch
import pandas as pd
import numpy as np
from typing import List, Dict


log = get_logger()

class EventExtractionDataModule(PLMBaseDataModule):
    """事件抽取数据模块

    数据集格式:
        {'text':'油服巨头哈里伯顿裁员650人 因美国油气开采活动放缓',
        'events':[{'label':'裁员事件','trigger':{'indices':[8,9], 'text':裁员'}, 'args':[{'label':'裁员方', 'indices':[4,5,6,7], 'text':'哈里伯顿'}]}]}
    """
    def __init__(self, 
                dataset: str, 
                plm: str, 
                batch_size: int,
                **kwargs):
        super().__init__()


    def setup(self, stage: str = 'fit') -> None:
        self.hparams.id2event = self.id2event
        self.hparams.id2arg = self.id2arg
        self.hparams.id2ent = self.id2ent
        self.hparams.id2combined = self.id2combined
        
            
            
    @property
    @lru_cache()
    def event_labels(self):
        labels = pd.Series(np.concatenate(self.train_df.events.values)).apply(lambda x: x['label']).drop_duplicates().values
        return labels.tolist()
    
    
    @property
    def event2id(self):
        return {l:i for i, l in enumerate(self.event_labels)}
    
    @property
    def id2event(self):
        return {i:l for l,i in self.event2id.items()}
    
    @property
    def trigger_labels(self):
        "将触发词改为 XX事件-触发词 这种样式"
        return [e + '-' + '触发词' for e in self.event_labels]
    
    
    @property
    @lru_cache()
    def arg_labels(self) -> Dict:
        labels = pd.Series(np.concatenate(pd.Series(np.concatenate(self.train_df.events.values)).apply(lambda x: x['args']).values)).apply(lambda x: x['label']).drop_duplicates().values
        return labels.tolist()
    
    
    @property
    def arg2id(self) -> Dict:
        return {l:i for i, l in enumerate(self.arg_labels)}
    
    @property
    def id2arg(self) -> Dict:
        return {i:l for l,i in self.arg2id.items()}
    
    @property
    def ent_labels(self) -> List:
        """将触发词与各个事件角色当做实体"""
        return sorted(self.trigger_labels + self.arg_labels)
    
    @property
    def ent2id(self) -> Dict:
        return {l:i for i,l in enumerate(self.ent_labels)}
    
    @property
    def id2ent(self) -> Dict:
        return {i:l for l,i in self.ent2id.items()}


    @property
    @lru_cache()
    def combined_labels(self) -> List:
        """将事件类型跟所有事件角色(包括触发词)拼接为一个标签
        - 例如: 裁员事件-触发词, 裁员事件-裁员方
        """
        def get_labels(e):
            labels = []
            label = e['label']
            labels.append((label, '触发词')) #触发词同样添加进去
            for arg in e['args']:
                pair_label = (label,arg['label'])
                if pair_label not in labels:
                    labels.append(pair_label)
            return labels
        df = pd.DataFrame(np.concatenate(pd.Series(np.concatenate(self.train_df.events.values)).apply(get_labels).values).tolist()).drop_duplicates()
        labels = sorted([tuple(v) for v in df.values])
        return labels


    @property
    def combined2id(self) -> Dict:
        return {l:i for i,l in enumerate(self.combined_labels)}
    
    @property
    def id2combined(self) -> Dict:
        return {i:l for i,l in enumerate(self.combined_labels)}


    def combined_transform(self, examples):
        batch_text = examples['text']
        batch_events = examples['events']
        batch_role_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      max_length=max_length,
                                      padding='max_length',
                                      truncation=True,
                                      return_offsets_mapping=True,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_mappings = batch_inputs.pop('offset_mapping').tolist()
        for i, text in enumerate(batch_text):
            offset_mapping = batch_mappings[i]
            events = batch_events[i]
            role_ids = torch.zeros(len(self.combined2id), max_length, max_length, dtype=torch.long)
            head_ids = torch.zeros(1, max_length, max_length, dtype=torch.long)
            tail_ids = torch.zeros(1, max_length, max_length, dtype=torch.long)
            for event in events:
                e_label = event['label']
                args = event['args']
                # 将触发词当做事件角色之一
                trigger = event['trigger']
                trigger_indices = trigger['indices']
                args.append({'label':'触发词', 'indices': trigger_indices, 'text': trigger})
                for i, arg1 in enumerate(args):
                    label = (e_label, arg1['label'])
                    arg1_head = arg1['indices'][0]
                    arg1_tail = arg1['indices'][-1]
                    try:
                        _arg1_head = char_idx_to_token(arg1_head, offset_mapping=offset_mapping)
                        _arg1_tail = char_idx_to_token(arg1_tail, offset_mapping=offset_mapping)
                        assert _arg1_head > 0
                        assert _arg1_tail > 0
                        role_ids[self.combined2id[label]][_arg1_head][_arg1_tail] = 1
                    except:
                        log.warning(f'batch max length: {max_length}, role {arg1["text"]} offset {(arg1_head, arg1_tail)} align to token offset failed in \n\t {text}')
                        continue
                        # 这个role 跟 其他的每个role 头头 尾尾 联系起来
                    for j, arg2 in enumerate(args):
                        if j>i:
                            arg2_head = arg2['indices'][0]
                            arg2_tail = arg2['indices'][-1]
                            try:
                                _arg2_head = char_idx_to_token(arg2_head, offset_mapping=offset_mapping)
                                _arg2_tail = char_idx_to_token(arg2_tail, offset_mapping=offset_mapping)
                                assert _arg2_head > 0
                                assert _arg2_tail > 0
                                head_ids[0][min(_arg1_head, _arg2_head)][max(_arg1_head, _arg2_head)] = 1
                                tail_ids[0][min(_arg1_tail, _arg2_tail)][max(_arg1_tail, _arg2_tail)] = 1
                            except:
                                log.warning(f'batch max length: {max_length}, role {arg2["text"]} offset {(arg2_head, arg2_tail)} align to token offset failed in \n\t {text}')
                                continue
            batch_role_tags.append(role_ids)
            batch_head_tags.append(head_ids)
            batch_tail_tags.append(tail_ids)
        batch_inputs['role_tags'] = torch.stack(batch_role_tags, dim=0)
        batch_inputs['head_tags'] = torch.stack(batch_head_tags, dim=0)
        batch_inputs['tail_tags'] = torch.stack(batch_tail_tags, dim=0)
        return batch_inputs
    
    def sparse_combined_transform(self, examples):
        batch_text = examples['text']
        batch_events = examples['events']
        batch_role_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        max_length = self.get_batch_max_length(batch_text=batch_text)
        batch_inputs = self.tokenizer(batch_text,
                                      max_length=max_length,
                                      padding='max_length',
                                      truncation=True,
                                      return_offsets_mapping=True,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_mappings = batch_inputs.pop('offset_mapping').tolist()
        for i, text in enumerate(batch_text):
            offset_mapping = batch_mappings[i]
            events = batch_events[i]
            role_tags = [set() for _ in range(len(self.combined2id))]
            head_tags = [set()]
            tail_tags = [set()]
            for event in events:
                e_label = event['label']
                args = event['args']
                # 将触发词当做事件角色之一
                trigger = event['trigger']
                trigger_indices = trigger['indices']
                args.append({'label':'触发词', 'indices': trigger_indices, 'text': trigger['text']})
                for i, arg1 in enumerate(args):
                    label = (e_label, arg1['label'])
                    arg1_head = arg1['indices'][0]
                    arg1_tail = arg1['indices'][-1]
                    assert arg1_tail >= arg1_head
                    try:
                        _arg1_head = char_idx_to_token(arg1_head, offset_mapping=offset_mapping)
                        _arg1_tail = char_idx_to_token(arg1_tail, offset_mapping=offset_mapping)
                        assert _arg1_head > 0
                        assert _arg1_tail > 0
                        role_tags[self.combined2id[label]].add((_arg1_head, _arg1_tail))
                    except:
                        log.warning(f'batch max length: {max_length}, arg1 {arg1["text"]} offset {(arg1_head, arg1_tail)} align to token offset failed in \n\t {text}')
                        continue
                        # 这个role 跟 其他的每个role 头头 尾尾 联系起来
                    for j, arg2 in enumerate(args):
                        if j>i:
                            arg2_head = arg2['indices'][0]
                            arg2_tail = arg2['indices'][-1]
                            try:
                                _arg2_head = char_idx_to_token(arg2_head, offset_mapping=offset_mapping)
                                _arg2_tail = char_idx_to_token(arg2_tail, offset_mapping=offset_mapping)
                                assert _arg2_head > 0
                                assert _arg2_tail > 0
                                head_tags[0].add((min(_arg1_head, _arg2_head), max(_arg1_head, _arg2_head)))
                                tail_tags[0].add((min(_arg1_tail, _arg2_tail), max(_arg1_tail, _arg2_tail)))
                            except:
                                log.warning(f'batch max length: {max_length}, arg2 {arg2["text"]} offset {(arg2_head, arg2_tail)} align to token offset failed in \n\t {text}')
                                continue
            for tag in role_tags + head_tags + tail_tags:
                if not tag:
                    tag.add((0,0))
            role_tags = sequence_padding([list(l) for l in role_tags])
            head_tags = sequence_padding([list(l) for l in head_tags])
            tail_tags = sequence_padding([list(l) for l in tail_tags])
            batch_role_tags.append(role_tags)
            batch_head_tags.append(head_tags)
            batch_tail_tags.append(tail_tags) 
        batch_inputs['role_tags'] = torch.tensor(sequence_padding(batch_role_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['head_tags'] = torch.tensor(sequence_padding(batch_head_tags, seq_dims=2), dtype=torch.long)
        batch_inputs['tail_tags'] = torch.tensor(sequence_padding(batch_tail_tags, seq_dims=2), dtype=torch.long)
        return batch_inputs
    
    def blinker_transform(self, examples):
        """论元角色 头相连,尾相连
        """
        batch_text = examples['text']
        batch_events = examples['events']
        batch_role_tags = []
        batch_head_tags = []
        batch_tail_tags = []
        max_length = self.get_max_length()
        batch_inputs = self.tokenizer(batch_text,
                                      max_length=max_length,
                                      padding='max_length',
                                      truncation=True,
                                      return_offsets_mapping=True,
                                      return_tensors='pt',
                                      return_token_type_ids=False)
        batch_mappings = batch_inputs.pop('offset_mapping').tolist()
        for i, text in enumerate(batch_text):
            offset_mapping = batch_mappings[i]
            events = batch_events[i]
            role_tags = torch.zeros(len(self.event_labels), max_length, max_length, dtype=torch.long) # +1, 加上触发词
            head_tags = torch.zeros(len(self.event_labels), max_length, max_length, dtype=torch.long)
            tail_tags = torch.zeros(len(self.event_labels), max_length, max_length, dtype=torch.long)
            for event in events:
                args: List = event['args']
                e_label = event['label']
                trigger = event['trigger']
                trigger_head = trigger['offset'][0]
                trigger_tail = trigger['offset'][1]
                args.append({'label':e_label + '-' + '触发词', 'offset': (trigger_head, trigger_tail)})
                for i, role1 in enumerate(args):
                    role1_head = role1['offset'][0]
                    role1_tail = role1['offset'][1]-1
                    try:
                        _role1_head = char_idx_to_token(role1_head, offset_mapping=offset_mapping)
                        _role1_tail = char_idx_to_token(role1_tail, offset_mapping=offset_mapping)
                        role_tags[self.event2id[e_label]][_role1_head][_role1_tail] = 1
                    except:
                        log.warning(f'role: {role1["text"]} offset {(role1_head, role1_tail)} align failed in \n\t text: {text}')
                        continue
                        # 这个role 跟 其他的每个role 头头 尾尾 联系起来
                    for j, role2 in enumerate(args):
                        if j>i:
                            role2_head = role2['offset'][0]
                            role2_tail = role2['offset'][1]-1
                            try:
                                _role2_head = char_idx_to_token(role2_head, offset_mapping=offset_mapping)
                                _role2_tail = char_idx_to_token(role2_tail, offset_mapping=offset_mapping)
                                head_tags[self.event2id[e_label]][min(_role1_head, _role2_head)][max(_role1_head, _role2_head)] = 1
                                tail_tags[self.event2id[e_label]][min(_role1_tail, _role2_tail)][max(_role1_tail, _role2_tail)] = 1
                            except:
                                log.warning(f'role: {role2["text"]} offset {(role2_head, role2_tail)} align failed \n\t text: {text}')
                                continue
            batch_role_tags.append(role_tags)
            batch_head_tags.append(head_tags)
            batch_tail_tags.append(tail_tags)
        batch_inputs['role_tags'] = torch.stack(batch_role_tags, dim=0)
        batch_inputs['head_tags'] = torch.stack(batch_head_tags, dim=0)
        batch_inputs['tail_tags'] = torch.stack(batch_tail_tags, dim=0)
        return batch_inputs


from ..utils.make_datamodule import align_char_span_text_b, PLMBaseDataModule
import torch
from typing import Dict, Union
import logging

log = logging.getLogger()


class PromptRelationExtractionDataModule(PLMBaseDataModule):
    def __init__(self,
                dataset: str,
                plm: str,
                auto_length: Union[str, int],
                transform: str,
                batch_size: int,
                pin_memory: bool=False,
                num_workers: int=0,
                dataset_dir: str ='./datasets/',
                plm_dir: str = './plms/') :
        """基于模板提示的文本片段抽取数据模块

        Args:
            dataset (str): 数据集名称
            plm (str): 预训练模型名称
            max_length (int): 文本最大长度
            batch_size (int): 批次大小
            pin_memory (bool, optional): 锁页内存. Defaults to True.
            num_workers (int, optional): 多进程. Defaults to 0.
            dataset_dir (str, optional): 数据集目录. Defaults to './datasets/'.
            plm_dir (str, optional): 预训练模型目录. Defaults to './plms/'.
        """
        super().__init__()
        self.transforms = {'prompt_gplinker': self.gplinker_transform}        
        assert self.hparams.transform in self.transforms.keys(), f'availabel models for relation extraction: {self.transforms.keys()}'
    
        
    def setup(self, stage: str) -> None:
        self.hparams.max_length = self.get_max_length()
        self.dataset.set_transform(transform=self.transforms.get(self.hparams.transform))
        
        
    def gplinker_transform(self, example) -> Dict:
        batch_text = example['text']
        batch_triples = example['triples']
        batch_prompts = example['prompts']
        batch_inputs = {'input_ids': [], 'attention_mask': [], 'token_type_ids': [], 'so_ids': [], 'head_ids': [], 'tail_ids': []}
        for i, text in enumerate(batch_text):
            prompt = batch_prompts[i]
            inputs = self.tokenizer(
                prompt,
                text, 
                padding='max_length',  
                max_length=self.hparams.max_length,
                truncation=True,
                return_offsets_mapping=True)
            batch_inputs['input_ids'].append(inputs['input_ids'])
            batch_inputs['attention_mask'].append(inputs['attention_mask'])
            batch_inputs['token_type_ids'].append(inputs['token_type_ids'])
            offset_mapping = inputs['offset_mapping']
            so_ids = torch.zeros(2, self.hparams.max_length, self.hparams.max_length)
            head_ids = torch.zeros(1, self.hparams.max_length, self.hparams.max_length)
            tail_ids = torch.zeros(1, self.hparams.max_length, self.hparams.max_length)
            triples = batch_triples[i]
            for triple in triples:
                #加1是因为有cls
                sub_start = triple['subject']['offset'][0] 
                sub_end = triple['subject']['offset'][1]
                obj_start = triple['object']['offset'][0] 
                obj_end = triple['object']['offset'][1]
                try:
                    sub_start_, sub_end_ = align_char_span_text_b((sub_start, sub_end), offset_mapping) 
                    obj_start_, obj_end_ = align_char_span_text_b((obj_start,obj_end), offset_mapping)
                    so_ids[0][sub_start_][sub_end_-1] = 1
                    so_ids[1][obj_start_][obj_end_-1] = 1
                    head_ids[0][sub_start_][obj_start_] = 1
                    tail_ids[0][sub_end_-1][obj_end_-1] = 1
                except Exception as e:
                    log.warning(e)
                    log.warning(f'sub char offset {(sub_start, sub_end)} or obj char offset {(obj_start, obj_end)} align to token offset failed \n text: {text} \n mapping: {offset_mapping}')
                    continue
            batch_inputs['so_ids'].append(so_ids)
            batch_inputs['head_ids'].append(head_ids)
            batch_inputs['tail_ids'].append(tail_ids)
        batch_inputs['so_ids'] = torch.stack(batch_inputs['so_ids'], dim=0)
        batch_inputs['head_ids'] = torch.stack(batch_inputs['head_ids'], dim=0)
        batch_inputs['tail_ids'] = torch.stack(batch_inputs['tail_ids'], dim=0)
        batch = dict(zip(batch_inputs.keys(), map(torch.tensor, batch_inputs.values())))
        return batch